# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

"""
Based on the mobilevit official implementation.
https://github.com/apple/ml-cvnets/blob/6acab5e446357cc25842a90e0a109d5aeeda002f/cvnets/models/classification/mobilevit.py
"""

import argparse
import math
from typing import Any, Dict, Literal, Optional, Tuple, Union, List

from omegaconf import DictConfig
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor

from ...op.base_metaformer import ChannelMLP, MetaFormer, MetaFormerBlock, MetaFormerEncoder, MultiHeadAttention
from ...op.custom import ConvLayer, GlobalPool, InvertedResidual
from ...utils import BackboneOutput, FXTensorType
from ..registry import USE_INTERMEDIATE_FEATURES_TASK_LIST

__all__ = ['mobilevit']
SUPPORTING_TASK = ['classification']


class MobileViTEmbeddings(nn.Module):
    def __init__(self, image_channels, hidden_size):
        super().__init__()
        
        image_channels = 3  # {RGB}
        self.conv = ConvLayer(
            in_channels=image_channels,
            out_channels=hidden_size,
            kernel_size=3,
            stride=2,
            use_norm=True,
            use_act=True,
            act_type='silu',
        )
    
    def forward(self, x):
        # x: B x 3(={RGB}) x H x W
        x = self.conv(x)  # B x C x H//2 x W//2
        return x


class MobileViTTransformerBlock(MetaFormerBlock):
    # Original: TransformerEncoder
    def __init__(self, hidden_size, num_attention_heads, attention_dropout_prob, intermediate_size, hidden_dropout_prob, layer_norm_eps) -> None:
        super().__init__(hidden_size, layer_norm_eps)
        self.layernorm_before = nn.LayerNorm(hidden_size, eps=layer_norm_eps)
        self.layernorm_after = nn.LayerNorm(hidden_size, eps=layer_norm_eps)
        self.token_mixer = MultiHeadAttention(hidden_size, num_attention_heads,
                                              attention_dropout_prob=attention_dropout_prob,
                                              use_qkv_bias=True)
        self.channel_mlp = ChannelMLP(hidden_size, intermediate_size, hidden_dropout_prob)


class MobileViTBlock(nn.Module):
    # Original: MobileViTBlock
    def __init__(self, num_transformer_blocks, in_channels, hidden_size, local_kernel_size,
                 patch_h, patch_w, num_attention_heads, attention_dropout_prob, intermediate_size, hidden_dropout_prob,
                 layer_norm_eps, use_fusion_layer=True) -> None:
        super().__init__()
        self.patch_h = patch_h
        self.patch_w = patch_w
        self.patch_area = self.patch_h * self.patch_w
        
        self.local_rep = nn.Sequential(*[
            ConvLayer(in_channels=in_channels, out_channels=in_channels,
                      kernel_size=local_kernel_size, stride=1, 
                      use_norm=True, use_act=True, act_type='silu'),
            ConvLayer(in_channels=in_channels, out_channels=hidden_size,
                      kernel_size=1, stride=1,
                      use_norm=False, use_act=False)
        ])
        
        global_rep_blocks = [
            MobileViTTransformerBlock(
                hidden_size=hidden_size,  # embed_dim, transformer_dim
                num_attention_heads=num_attention_heads,  # num_heads
                attention_dropout_prob=attention_dropout_prob,  # attn_dropout
                intermediate_size=intermediate_size,  # ffn_latent_dim, ffn_dim
                hidden_dropout_prob=hidden_dropout_prob,  # ffn_dropout
                layer_norm_eps=layer_norm_eps
            )
            for _ in range(num_transformer_blocks)
        ]
        global_rep_blocks.append(nn.LayerNorm(normalized_shape=hidden_size))
        self.global_rep = nn.Sequential(*global_rep_blocks)
        
        self.proj = ConvLayer(in_channels=hidden_size, out_channels=in_channels,
                              kernel_size=1, stride=1,
                              use_norm=True, use_act=True, act_type='silu')
        
        self.use_fusion_layer = use_fusion_layer
        if self.use_fusion_layer:
            self.fusion = ConvLayer(in_channels=in_channels * 2, out_channels=in_channels,
                                    kernel_size=local_kernel_size, stride=1,
                                    use_norm=True, use_act=True, act_type='silu')
        
    def unfolding(self, feature_map: Tensor) -> Tuple[Tensor, Dict]:
        batch_size, in_channels, orig_h, orig_w = feature_map.size()

        # @deepkyu: [fx tracing] remove ceil becuase mod == 0, while ceil is not supported mostly in fx
        # new_h = math.ceil(orig_h / self.patch_h) * self.patch_h
        # new_w = math.ceil(orig_w / self.patch_w) * self.patch_w
        
        new_h = orig_h // self.patch_h * self.patch_h
        new_w = orig_w // self.patch_w * self.patch_w

        interpolate = False
        # @deepkyu: [fx tracing] Found always satisfied: (new_w == orig_w) and (new_h == orig_h)
        # if new_w != orig_w or new_h != orig_h:
        #     # Note: Padding can be done, but then it needs to be handled in attention function.
        #     feature_map = F.interpolate(
        #         feature_map, size=(new_h, new_w), mode="bilinear", align_corners=False
        #     )
        #     interpolate = True

        # number of patches along width and height
        num_patch_w = new_w // self.patch_w  # n_w
        num_patch_h = new_h // self.patch_h  # n_h
        num_patches = num_patch_h * num_patch_w  # N

        # [B, C, H, W] --> [B, C, n_h, p_h, n_w, p_w]
        reshaped_fm = feature_map.reshape(
            batch_size, in_channels, num_patch_h, self.patch_h, num_patch_w, self.patch_w
        )
        # [B, C, n_h, p_h, n_w, p_w] --> [B, C, n_h, n_w, p_h, p_w]
        transposed_fm = reshaped_fm.transpose(3, 4)
        # [B, C, n_h, n_w, p_h, p_w] --> [B, C, N, P] where P = p_h * p_w and N = n_h * n_w
        reshaped_fm = transposed_fm.reshape(
            batch_size, in_channels, num_patches, self.patch_area
        )
        # [B, C, N, P] --> [B, P, N, C]
        transposed_fm = reshaped_fm.transpose(1, 3)
        # [B, P, N, C] --> [BP, N, C]
        patches = transposed_fm.reshape(batch_size * self.patch_area, num_patches, -1)

        info_dict = {
            "orig_size": (orig_h, orig_w),
            "batch_size": batch_size,
            "interpolate": interpolate,
            "total_patches": num_patches,
            "num_patches_w": num_patch_w,
            "num_patches_h": num_patch_h,
        }

        return patches, info_dict

    def folding(self, patches: Tensor, info_dict: Dict) -> Tensor:
        patches.dim()
        
        # @deepkyu: [fx tracing] Found always satisfied: assert n_dim == 3
        # assert n_dim == 3, "Tensor should be of shape BPxNxC. Got: {}".format(
        #     patches.shape
        # )
        
        # [BP, N, C] --> [B, P, N, C]
        patches = patches.contiguous().view(
            info_dict["batch_size"], self.patch_area, info_dict["total_patches"], -1
        )

        batch_size, pixels, num_patches, channels = patches.size()
        num_patch_h = info_dict["num_patches_h"]
        num_patch_w = info_dict["num_patches_w"]

        # [B, P, N, C] --> [B, C, N, P]
        patches = patches.transpose(1, 3)

        # [B, C, N, P] --> [B, C, n_h, n_w, p_h, p_w]
        feature_map = patches.reshape(
            batch_size, channels, num_patch_h, num_patch_w, self.patch_h, self.patch_w
        )
        # [B, C, n_h, n_w, p_h, p_w] --> [B, C, n_h, p_h, n_w, p_w]
        feature_map = feature_map.transpose(3, 4)
        # [B, C, n_h, p_h, n_w, p_w] --> [B, C, H, W]
        feature_map = feature_map.reshape(
            batch_size, channels, num_patch_h * self.patch_h, num_patch_w * self.patch_w
        )
        if info_dict["interpolate"]:
            feature_map = F.interpolate(
                feature_map,
                size=info_dict["orig_size"],
                mode="bilinear",
                align_corners=False,
            )
        return feature_map

    # def forward_spatial(self, x: Tensor) -> Tensor:
    #     res = x

    #     fm = self.local_rep(x)

    #     # convert feature map to patches
    #     patches, info_dict = self.unfolding(fm)

    #     # learn global representations
    #     for transformer_layer in self.global_rep:
    #         patches = transformer_layer(patches)

    #     # [B x Patch x Patches x C] --> [B x C x Patches x Patch]
    #     fm = self.folding(patches=patches, info_dict=info_dict)

    #     fm = self.conv_proj(fm)

    #     if self.fusion is not None:
    #         fm = self.fusion(torch.cat((res, fm), dim=1))
    #     return fm

    # def forward_temporal(
    #     self, x: Tensor, x_prev: Optional[Tensor] = None
    # ) -> Union[Tensor, Tuple[Tensor, Tensor]]:

    #     res = x
    #     fm = self.local_rep(x)

    #     # convert feature map to patches
    #     patches, info_dict = self.unfolding(fm)

    #     # learn global representations
    #     for global_layer in self.global_rep:
    #         if isinstance(global_layer, TransformerEncoder):
    #             patches = global_layer(x=patches, x_prev=x_prev)
    #         else:
    #             patches = global_layer(patches)

    #     # [B x Patch x Patches x C] --> [B x C x Patches x Patch]
    #     fm = self.folding(patches=patches, info_dict=info_dict)

    #     fm = self.conv_proj(fm)

    #     if self.fusion is not None:
    #         fm = self.fusion(torch.cat((res, fm), dim=1))
    #     return fm, patches

    def forward(
        self, x: Union[Tensor, Tuple[Tensor]], *args, **kwargs
    ) -> Union[Tensor, Tuple[Tensor, Tensor]]:

        out = self.local_rep(x)

        # convert feature map to patches
        patches, info_dict = self.unfolding(out)

        # learn global representations
        patches = self.global_rep(patches)

        # [B x Patch x Patches x C] --> [B x C x Patches x Patch]
        out = self.folding(patches=patches, info_dict=info_dict)

        out = self.proj(out)

        if self.use_fusion_layer:
            out = self.fusion(torch.cat((x, out), dim=1))  # skip concat
        return out


class MobileViTEncoder(MetaFormerEncoder):
    def __init__(
        self,
        params: Optional[DictConfig] = None,
        stage_params: Optional[List] = None,
    ) -> None:
        super().__init__()
        stages = []
        
        self.dilation = 1
        self.patch_size = params.patch_size
        self.num_attention_heads = params.num_attention_heads
        self.attention_dropout_prob = params.attention_dropout_prob
        self.hidden_dropout_prob = params.ffn_dropout_prob
        self.use_fusion_layer = params.use_fusion_layer

        # Fix as constant
        self.layer_norm_eps = 1e-5
        self.local_kernel_size = 3

        # Fix as constant
        in_channels = 16
        for stage_param in stage_params:
            stages.append(self._make_block(stage_param, in_channels))
            in_channels = stage_param.out_channels
        self.blocks = nn.Sequential(*stages)
    
    def _make_block(self, stage_param, in_channels):
        out_channels = stage_param.out_channels
        block_type = stage_param.block_type
        stride = stage_param.stride
        num_blocks = stage_param.num_blocks
        expand_ratio = stage_param.ir_expansion_ratio
        
        #out_channels, block_type: Literal['mv2', 'mobilevit'], num_blocks, stride, hidden_size, intermediate_size, num_transformer_blocks, dilate, expand_ratio, in_channels
        if block_type == 'mobilevit':
            num_transformer_blocks = num_blocks
            hidden_size = stage_param.attention_channels
            intermediate_size = stage_param.ffn_intermediate_channels
            dilate = stage_param.dilate

            return self._make_mobilevit_blocks(
                num_transformer_blocks, in_channels, out_channels, stride, expand_ratio,
                hidden_size, intermediate_size, self.patch_size, self.local_kernel_size, self.num_attention_heads,
                self.attention_dropout_prob, self.hidden_dropout_prob, self.layer_norm_eps, self.use_fusion_layer,
                dilate
            )
            
        return self._make_inverted_residual_blocks(
            num_blocks, in_channels, out_channels, stride, expand_ratio
        )
    
    def _make_inverted_residual_blocks(self, num_blocks, in_channels, out_channels, stride, expand_ratio):
        blocks = [
            InvertedResidual(
                in_channels=in_channels if block_idx == 0 else out_channels,
                hidden_channels=in_channels*expand_ratio if block_idx == 0 else out_channels*expand_ratio,
                out_channels=out_channels,
                kernel_size=3,
                stride=stride if block_idx == 0 else 1,
                dilation=1,
                act_type='silu',
            ) for block_idx in range(num_blocks)
        ]
        return nn.Sequential(*blocks)
    
    def _make_mobilevit_blocks(self, num_transformer_blocks, in_channels, out_channels, stride, expand_ratio,
                               hidden_size, intermediate_size, patch_size, local_kernel_size, num_attention_heads, attention_dropout_prob,
                               hidden_dropout_prob, layer_norm_eps, use_fusion_layer, dilate: Optional[bool] = False):
        blocks = []
        if stride == 2:
            blocks.append(
                InvertedResidual(
                    in_channels=in_channels,
                    hidden_channels=in_channels*expand_ratio,
                    out_channels=out_channels,
                    kernel_size=3,
                    stride=stride if not dilate else 1,
                    dilation=self.dilation,
                    act_type='silu',
                )
            )
            if dilate:
                self.dilation += 2
                
            in_channels = out_channels
            
        blocks.append(
            MobileViTBlock(
                num_transformer_blocks=num_transformer_blocks,
                in_channels=in_channels,
                hidden_size=hidden_size,
                local_kernel_size=local_kernel_size,  # getattr(opts, "model.classification.mit.conv_kernel_size", 3)
                patch_h=patch_size,  # patch_h
                patch_w=patch_size,  # patch_w
                num_attention_heads=num_attention_heads,  # num_heads
                attention_dropout_prob=attention_dropout_prob,  # attn_dropout=getattr(opts, "model.classification.mit.attn_dropout", 0.1)
                intermediate_size=intermediate_size,  # ffn_dim
                hidden_dropout_prob=hidden_dropout_prob,  # dropout, ffn_dropout
                layer_norm_eps=layer_norm_eps,
                use_fusion_layer=use_fusion_layer  # not getattr(opts, "model.classification.mit.no_fuse_local_global_features", False)
            )
        )
        
        return nn.Sequential(*blocks)
    

class MobileViT(MetaFormer):
    def __init__(
        self,
        task: str,
        params: Optional[DictConfig] = None,
        stage_params: Optional[List] = None,
    ) -> None:
        # Check task compatibility
        self.task = task.lower()
        assert self.task in SUPPORTING_TASK, f'MobileViT is not supported on {self.task} task now.'
        #self.use_intermediate_features = self.task in USE_INTERMEDIATE_FEATURES_TASK_LIST

        exp_channels = min(params.output_expansion_ratio * stage_params[-1].out_channels, 960)
        hidden_sizes = [stage.out_channels for stage in stage_params] + [exp_channels]
        super().__init__(hidden_sizes)
        
        self.task = task
        self.intermediate_features = self.task in ['segmentation', 'detection']
        
        image_channels = 3
        # Fix as constant
        patch_embedding_out_channels = 16
        self.patch_embed = MobileViTEmbeddings(image_channels, patch_embedding_out_channels)
        self.encoder = MobileViTEncoder(params=params, stage_params=stage_params)
        
        self.conv_1x1_exp = ConvLayer(in_channels=stage_params[-1].out_channels, out_channels=exp_channels,
                                      kernel_size=1, stride=1,
                                      use_act=True, use_norm=True, act_type='silu')
        self.pool = GlobalPool(pool_type="mean", keep_dim=False)
        
        self._feature_dim = exp_channels
        
    def forward(self, x: FXTensorType):
        x = self.patch_embed(x)
        x = self.encoder(x)
        x = self.conv_1x1_exp(x)
        feat = self.pool(x)
        return BackboneOutput(last_feature=feat)


def mobilevit(task, conf_model_backbone):
    return MobileViT(task, conf_model_backbone.params, conf_model_backbone.stage_params)



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import List

import numpy as np
import torch

from ..base import BaseMetric


@torch.no_grad()
def accuracy_topk(pred, target):
    """Computes the accuracy over the k top predictions for the specified values of k"""
    pred = pred.T
    class_num = pred.shape[0]
    correct = np.equal(pred, np.tile(target, (class_num, 1)))
    return correct


class ClassificationMetricAdaptor:
    '''
        Adapter to process redundant operations for the metrics.
    '''
    def __init__(self, metric_names) -> None:
        self.metric_names = metric_names

    def __call__(self, predictions: List[dict], targets: List[dict]):
        ret = {}
        if 'top1_accuracy' in self.metric_names or 'top5_accuracy' in self.metric_names:
            correct = accuracy_topk(predictions, targets)
            ret['correct'] = correct

        return ret

class Top1Accuracy(BaseMetric):
    def __init__(self, num_classes, classwise_analysis, **kwargs):
        metric_name = 'Acc@1' # Name for logging
        super().__init__(metric_name=metric_name, num_classes=num_classes, classwise_analysis=classwise_analysis)

    def calibrate(self, pred, target, **kwargs):
        topk = 1
        correct = kwargs['correct']
        correct = np.logical_or.reduce(correct[:topk])

        if self.classwise_analysis: # Classwise analysis
            for correct_class, c in zip(target, correct):
                count = 1 if c else 0
                self.classwise_metric_meters[correct_class].update(count, n=1)

        Acc1_correct = correct.sum()
        self.metric_meter.update(Acc1_correct, n=pred.shape[0])


class Top5Accuracy(BaseMetric):
    def __init__(self, num_classes, classwise_analysis, **kwargs):
        metric_name = 'Acc@5' # Name for logging
        super().__init__(metric_name=metric_name, num_classes=num_classes, classwise_analysis=classwise_analysis)

    def calibrate(self, pred, target, **kwargs):
        topk = 5
        correct = kwargs['correct']
        correct = np.logical_or.reduce(correct[:topk])

        if self.classwise_analysis: # Classwise analysis
            for correct_class, c in zip(target, correct):
                count = 1 if c else 0
                self.classwise_metric_meters[correct_class].update(count, n=1)

        Acc1_correct = correct.sum()
        self.metric_meter.update(Acc1_correct, n=pred.shape[0])



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

DEFAULT_CROP_PCT = 0.95 #0.875
IMAGENET_DEFAULT_MEAN = (0.485, 0.456, 0.406)
IMAGENET_DEFAULT_STD = (0.229, 0.224, 0.225)
IMAGENET_INCEPTION_MEAN = (0.5, 0.5, 0.5)
IMAGENET_INCEPTION_STD = (0.5, 0.5, 0.5)
IMAGENET_DPN_MEAN = (124 / 255, 117 / 255, 104 / 255)
IMAGENET_DPN_STD = tuple([1 / (.0167 * 255)] * 3)

IMG_EXTENSIONS = ('.png', '.jpg', '.jpeg',
                  '.bmp', '.tif', '.tiff', '.dng', '.webp', '.mpo')



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import csv
import json
import re
from itertools import repeat
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import numpy as np


def read_json(json_path):
    with open(json_path, 'r') as f:
        data = json.load(f)
    return data


def expand_to_chs(x, n):
    if not isinstance(x, (tuple, list)):
        x = tuple(repeat(x, n))
    elif len(x) == 1:
        x = x * n
    else:
        assert len(x) == n, 'normalization stats must match image channels'
    return x


def natural_key(string_):
    """See http://www.codinghorror.com/blog/archives/001018.html"""
    return [int(s) if s.isdigit() else s for s in re.split(r'(\d+)', string_.lower())]


def get_detection_label(label_file: Path):
    target = Path(label_file).read_text()

    if target == '': # target label can be empty string
        target_array = np.zeros((0, 5))
    else:
        try:
            target_array = np.array([list(map(float, box.split(' '))) for box in target.split('\n') if box.strip()])
        except ValueError as e:
            print(target)
            raise e

    label, boxes = target_array[:, 0], target_array[:, 1:]
    label = label[..., np.newaxis]
    return label, boxes


def load_classification_class_map(labels_path: Optional[Union[str, Path]]):
    # Assume the `map_or_filename` is path for csv label file
    assert labels_path.exists(), f"Cannot locate specified class map file {labels_path}!"
    class_map_ext = labels_path.suffix.lower()
    assert class_map_ext == '.csv', f"Unsupported class map file extension ({class_map_ext})!"

    with open(labels_path, newline='') as csvfile:
        reader = csv.DictReader(csvfile)
        file_to_idx = {row['image_id']: int(row['class']) for row in reader}

    return file_to_idx


def is_file_dict(image_dir: Union[Path, str], file_or_dir_to_idx):
    image_dir = Path(image_dir)
    candidate_name = list(file_or_dir_to_idx.keys())[0]
    file_or_dir: Path = image_dir / candidate_name
    if file_or_dir.exists():
        return file_or_dir.is_file()

    file_candidates = list(image_dir.glob(f"{candidate_name}.*"))
    assert len(file_candidates) != 0, f"Unknown label format! Is there any something file like {file_or_dir} ?"

    return True


def as_tuple(tuple_string: str) -> Tuple:
    tuple_string = tuple_string.replace("(", "")
    tuple_string = tuple_string.replace(")", "")
    tuple_value = tuple(map(int, tuple_string.split(",")))
    return tuple_value



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from functools import partial

import torch
import torch.nn.functional as F
import torchvision
from torchvision.models.detection._utils import BoxCoder, _topk_min
from torchvision.ops import boxes as box_ops

from netspresso_trainer.utils.bbox_utils import generate_anchors, transform_bbox

from ..models.op.custom import Anchor2Vec
from ..models.utils import ModelOutput


def rtdetr_decode(pred, original_shape, num_top_queries=300, score_thresh=0.0):
    pred = pred['pred']
    boxes, logits = pred[..., :4], pred[..., 4:]

    num_classes = logits.shape[-1]
    h, w = original_shape[1], original_shape[2]
    boxes = transform_bbox(boxes, "cxcywhn -> xyxy", image_shape=(h, w))

    scores = torch.sigmoid(logits)
    scores, index = torch.topk(scores.flatten(1), num_top_queries, axis=-1)
    labels = index % num_classes
    index = index // num_classes
    boxes = boxes.gather(dim=1, index=index.unsqueeze(-1).repeat(1, 1, boxes.shape[-1]))

    # x1, y1, x2, y2, obj_conf, pred_score, pred_label
    detections = []
    for box, score, label in zip(boxes, scores, labels):
        keep = score > score_thresh
        # TODO: Dummy obj_conf should be removed
        detections.append(torch.cat([box[keep], torch.ones_like(score[keep].unsqueeze(-1)), score[keep].unsqueeze(-1), label[keep].unsqueeze(-1)], dim=-1))

    return detections


def anchor_decoupled_head_decode(pred, original_shape, topk_candidates=1000, score_thresh=0.05):
    box_coder = BoxCoder(weights=(1.0, 1.0, 1.0, 1.0))

    class_logits = pred["cls_logits"]
    box_regression = pred["bbox_regression"]
    anchors = pred["anchors"]

    detections = []

    for index in range(len(box_regression[0])):
        box_regression_per_image = [br[index] for br in box_regression]
        logits_per_image = [cl[index] for cl in class_logits]
        anchors_per_image = anchors.split([b.shape[0] for b in box_regression_per_image])

        image_boxes = []
        image_scores = []
        image_labels = []

        for box_regression_per_level, logits_per_level, anchors_per_level in zip(
            box_regression_per_image, logits_per_image, anchors_per_image
        ):
            num_classes = logits_per_level.shape[-1]

            # remove low scoring boxes
            scores_per_level = torch.sigmoid(logits_per_level).flatten()
            keep_idxs = scores_per_level > score_thresh
            scores_per_level = scores_per_level[keep_idxs]
            topk_idxs = torch.where(keep_idxs)[0]

            # keep only topk scoring predictions
            num_topk = _topk_min(topk_idxs, topk_candidates, 0)
            scores_per_level, idxs = scores_per_level.topk(num_topk)
            topk_idxs = topk_idxs[idxs]

            anchor_idxs = torch.div(topk_idxs, num_classes, rounding_mode="floor")
            labels_per_level = topk_idxs % num_classes

            boxes_per_level = box_coder.decode_single(
                box_regression_per_level[anchor_idxs], anchors_per_level[anchor_idxs]
            )
            boxes_per_level = box_ops.clip_boxes_to_image(boxes_per_level, original_shape[1:])

            image_boxes.append(boxes_per_level)
            image_scores.append(scores_per_level)
            image_labels.append(labels_per_level)

        image_boxes = torch.cat(image_boxes, dim=0)
        image_scores = torch.cat(image_scores, dim=0)
        image_labels = torch.cat(image_labels, dim=0).to(image_scores.dtype)
        tmp_obj = torch.ones(image_labels.shape, dtype=image_labels.dtype, device=image_labels.device).view(-1, 1)

        # x1, y1, x2, y2, dummy, pred_score, pred_label
        detections.append(
            torch.cat([image_boxes, tmp_obj.view(-1, 1), image_scores.view(-1, 1), image_labels.view(-1, 1)], dim=1)
        )

    return detections


def anchor_free_decoupled_head_decode(pred, original_shape, score_thresh=0.7):
    pred = pred['pred']
    dtype = pred[0].type()
    stage_strides= [original_shape[-1] // o.shape[-1] for o in pred]

    hw = [x.shape[-2:] for x in pred]
    # [batch, n_anchors_all, num_classes + 5]
    pred = torch.cat([x.flatten(start_dim=2) for x in pred], dim=2).permute(0, 2, 1)
    pred[..., 4:] = pred[..., 4:].sigmoid()

    grids = []
    strides = []
    for (hsize, wsize), stride in zip(hw, stage_strides):
        yv, xv = torch.meshgrid(torch.arange(hsize), torch.arange(wsize), indexing='ij')
        grid = torch.stack((xv, yv), 2).view(1, -1, 2)
        grids.append(grid)
        shape = grid.shape[:2]
        strides.append(torch.full((*shape, 1), stride))

    grids = torch.cat(grids, dim=1).type(dtype)
    strides = torch.cat(strides, dim=1).type(dtype)

    pred = torch.cat([
        (pred[..., 0:2] + grids) * strides,
        torch.exp(pred[..., 2:4]) * strides,
        pred[..., 4:]
    ], dim=-1)

    box_corner = pred.new(pred.shape)
    box_corner[:, :, :4] = transform_bbox(pred[:, :, :4], "cxcywh -> xyxy")
    pred[:, :, :4] = box_corner[:, :, :4]

    # Discard boxes with low score
    detections = []
    for p in pred:
        class_conf, class_pred = torch.max(p[:, 5:], 1, keepdim=True)

        conf_mask = (p[:, 4] * class_conf.squeeze() >= score_thresh).squeeze()

        # x1, y1, x2, y2, obj_conf, pred_score, pred_label
        detections.append(
            torch.cat((p[:, :5], class_conf, class_pred.float()), 1)[conf_mask]
        )

    return detections


def yolo_fastest_head_decode(pred, original_shape, score_thresh=0.7, anchors=None):
    pred = pred['pred']
    dtype = pred[0].type()
    stage_strides = [original_shape[-1] // o.shape[-1] for o in pred]
    hw = [x.shape[-2:] for x in pred]
    device = pred[0].device
    anchors = [torch.tensor(anchor, dtype=torch.float).view(-1, 2) for anchor in anchors]
    num_anchors = anchors[0].shape[0]
    anchors = torch.stack(anchors, dim=0).to(device)

    grids = []
    strides = []
    for (hsize, wsize), stride in zip(hw, stage_strides):
        yv, xv = torch.meshgrid(torch.arange(hsize), torch.arange(wsize), indexing='ij')
        grid = torch.stack((xv, yv), 2).repeat(num_anchors, 1,1,1).view(1, num_anchors, hsize, wsize, 2).type(dtype).to(device)
        grids.append(grid)
        shape = grid.shape[:-1]
        strides.append(torch.full((*shape, 1), stride).to(device))

    preds = []
    for idx, p in enumerate(pred):
        p = p.permute(0, 1, 3, 4, 2)
        p = torch.cat([
            (p[..., 0:2].sigmoid() + grids[idx]) * strides[idx],
            2. * (torch.tanh(p[..., 2:4]/2 -.549306) + 1.) * anchors[idx].view(1, num_anchors, 1, 1, 2),
            p[..., 4:].sigmoid()
        ], dim=-1).flatten(start_dim=1, end_dim=-2)
        preds.append(p)
    pred = torch.cat(preds, dim=1)

    box_corner = pred.new(pred.shape)
    box_corner[:, :, :4] = transform_bbox(pred[:, :, :4], "cxcywh -> xyxy")
    pred[:, :, :4] = box_corner[:, :, :4]

    # Discard boxes with low score
    detections = []
    for p in pred:
        class_conf, class_pred = torch.max(p[:, 5:], 1, keepdim=True)

        conf_mask = (p[:, 4] * class_conf.squeeze() >= score_thresh).squeeze()

        # x1, y1, x2, y2, obj_conf, pred_score, pred_label
        detections.append(
            torch.cat((p[:, :5], class_conf, class_pred.float()), 1)[conf_mask]
        )

    return detections

def yolo_head_decode(pred, original_shape, score_thresh=0.7, anc2vec=None, reg_max=16):
    pred = pred['pred']
    if isinstance(pred, dict):
        pred = pred['outputs']
    h, w = original_shape[1], original_shape[2]
    device = pred[0][0].device
    stage_strides= [original_shape[-1] // o.shape[-1] for o in pred]
    offset, scaler = generate_anchors((h, w), stage_strides)
    offset = offset.to(device)
    scaler = scaler.to(device)

    pred_bbox_reg, pred_class_logits = [], []
    for layer_output in pred:
        layer_output = layer_output.float()
        reg, class_logits = torch.split(layer_output, [4 * reg_max, layer_output.shape[1] - 4 * reg_max], dim=1)
        _, bbox_reg = anc2vec(reg)
        b, c, h, w = bbox_reg.shape
        reg = bbox_reg.permute(0, 2, 3, 1).view(b, h*w, c)
        pred_bbox_reg.append(reg)

        b, c, h, w = class_logits.shape
        logits = class_logits.permute(0, 2, 3, 1).view(b, h*w, c)
        pred_class_logits.append(logits)

    pred_bbox_reg = torch.concat(pred_bbox_reg, dim=1)
    pred_class_logits = torch.concat(pred_class_logits, dim=1).sigmoid()

    pred_xyxy = pred_bbox_reg * scaler.view(1, -1, 1)
    lt, rb = pred_xyxy.chunk(2, dim=-1)
    pred_bbox_reg = torch.cat([offset - lt, offset + rb], dim=-1)

    detections = []
    for bbox, cls_logits in zip(pred_bbox_reg, pred_class_logits):
        class_conf, class_pred = torch.max(cls_logits, 1, keepdim=True)
        conf_mask = (class_conf.squeeze() >= score_thresh).squeeze()

        detections.append(
            torch.cat((bbox, torch.ones_like(class_pred), class_conf, class_pred.float()), 1)[conf_mask]
        )
    return detections

def nms(prediction, nms_thresh=0.45, class_agnostic=False):
    output = [torch.zeros(0, 7).to(prediction[0].device) for i in range(len(prediction))]
    for i, image_pred in enumerate(prediction):

        # If none are remaining => process next image
        if not image_pred.size(0):
            continue

        if class_agnostic:
            nms_out_index = torchvision.ops.nms(
                image_pred[:, :4],
                image_pred[:, 4] * image_pred[:, 5],
                nms_thresh,
            )
        else:
            nms_out_index = torchvision.ops.batched_nms(
                image_pred[:, :4],
                image_pred[:, 4] * image_pred[:, 5],
                image_pred[:, 6],
                nms_thresh,
            )

        image_pred = image_pred[nms_out_index]
        output[i] = torch.cat((output[i], image_pred))

    return output


class DetectionPostprocessor:
    def __init__(self, conf_model):
        head_name = conf_model.architecture.head.name
        params = conf_model.postprocessor.params
        if head_name == 'anchor_free_decoupled_head':
            self.decode_outputs = partial(anchor_free_decoupled_head_decode, score_thresh=params.score_thresh)
            self.postprocess = partial(nms, nms_thresh=params.nms_thresh, class_agnostic=params.class_agnostic)
        elif head_name == 'anchor_decoupled_head':
            self.decode_outputs = partial(anchor_decoupled_head_decode, topk_candidates=params.topk_candidates, score_thresh=params.score_thresh)
            self.postprocess = partial(nms, nms_thresh=params.nms_thresh, class_agnostic=params.class_agnostic)
        elif head_name == 'yolo_fastest_head_v2':
            self.decode_outputs = partial(yolo_fastest_head_decode, score_thresh=params.score_thresh, anchors=params.anchors)
            self.postprocess = partial(nms, nms_thresh=params.nms_thresh, class_agnostic=params.class_agnostic)
        elif head_name == 'yolo_detection_head':
            self.anc2vec = Anchor2Vec(params.reg_max)
            self.decode_outputs = partial(yolo_head_decode, score_thresh=params.score_thresh, anc2vec=self.anc2vec, reg_max=params.reg_max)
            self.postprocess = partial(nms, nms_thresh=params.nms_thresh, class_agnostic=params.class_agnostic)
        elif head_name == 'rtdetr_head':
            self.decode_outputs = partial(rtdetr_decode, num_top_queries=params.num_top_queries, score_thresh=params.score_thresh)
            self.postprocess = None
        else:
            self.decode_outputs = None
            self.postprocess = None

    def __call__(self, outputs: ModelOutput, original_shape):
        pred = outputs

        if self.decode_outputs:
            pred = self.decode_outputs(pred, original_shape)
        if self.postprocess:
            pred = self.postprocess(pred)

        pred = [(torch.cat([p[:, :4], p[:, 4:5] * p[:, 5:6]], dim=-1).detach().cpu().numpy(),
                      p[:, 6].to(torch.int).detach().cpu().numpy())
                      for p in pred]
        return pred



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import Literal, Optional

import numpy as np
import torch

from .base import BaseTaskProcessor


class ClassificationProcessor(BaseTaskProcessor):
    def __init__(self, conf, postprocessor, devices, **kwargs):
        super(ClassificationProcessor, self).__init__(conf, postprocessor, devices, **kwargs)

    def train_step(self, train_model, batch, optimizer, loss_factory, metric_factory):
        train_model.train()
        indices, images, labels = batch
        images = images.to(self.devices).to(self.data_type)
        labels = labels.to(self.devices).to(self.data_type)
        target = {'target': labels}

        optimizer.zero_grad()

        with torch.cuda.amp.autocast(enabled=self.mixed_precision):
            if isinstance(train_model, torch.nn.parallel.DistributedDataParallel):
                if hasattr(train_model.module.head, "_training_targets"): # for RT-DETR
                    train_model.module.head.set_training_targets(target)
            else:
                if hasattr(train_model.head, "_training_targets"):
                    train_model.head.set_training_targets(target)
            out = train_model(images)
            loss_factory.calc(out, target, phase='train')
        if labels.dim() > 1: # Soft label to label number
            labels = torch.argmax(labels, dim=-1)
        pred, _ = self.postprocessor(out)

        loss_factory.backward(self.grad_scaler)
        if self.max_norm:
            # Unscales the gradients of optimizer's assigned parameters in-place
            self.grad_scaler.unscale_(optimizer)
            torch.nn.utils.clip_grad_norm_(train_model.parameters(), self.max_norm)
        # optimizer's gradients are already unscaled, so scaler.step doesn't unscale them,
        self.grad_scaler.step(optimizer)
        self.grad_scaler.update()

        labels = labels.detach().cpu().numpy() # Change it to numpy before compute metric
        if self.conf.distributed:
            gathered_pred = [None for _ in range(torch.distributed.get_world_size())]
            gathered_labels = [None for _ in range(torch.distributed.get_world_size())]

            torch.distributed.gather_object(pred, gathered_pred if torch.distributed.get_rank() == 0 else None, dst=0)
            torch.distributed.gather_object(labels, gathered_labels if torch.distributed.get_rank() == 0 else None, dst=0)
            torch.distributed.barrier()
            if torch.distributed.get_rank() == 0:
                [metric_factory.update(g_pred, g_labels, phase='train') for g_pred, g_labels in zip(gathered_pred, gathered_labels)]
        else:
            metric_factory.update(pred, labels, phase='train')

    def valid_step(self, eval_model, batch, loss_factory, metric_factory):
        eval_model.eval()
        indices, images, labels = batch
        images = images.to(self.devices)
        labels = labels.to(self.devices)
        target = {'target': labels}

        out = eval_model(images)
        loss_factory.calc(out, target, phase='valid')
        if labels.dim() > 1: # Soft label to label number
            labels = torch.argmax(labels, dim=-1)
        pred, conf_score = self.postprocessor(out)

        indices = indices.numpy()
        labels = labels.detach().cpu().numpy() # Change it to numpy before compute metric
        if self.conf.distributed:
            gathered_pred = [None for _ in range(torch.distributed.get_world_size())]
            gathered_labels = [None for _ in range(torch.distributed.get_world_size())]

            # Remove dummy samples, they only come in distributed environment
            pred = pred[indices != -1]
            labels = labels[indices != -1]
            torch.distributed.gather_object(pred, gathered_pred if torch.distributed.get_rank() == 0 else None, dst=0)
            torch.distributed.gather_object(labels, gathered_labels if torch.distributed.get_rank() == 0 else None, dst=0)
            torch.distributed.barrier()
            if torch.distributed.get_rank() == 0:
                [metric_factory.update(g_pred, g_labels, phase='valid') for g_pred, g_labels in zip(gathered_pred, gathered_labels)]
        else:
            metric_factory.update(pred, labels, phase='valid')

        if self.single_gpu_or_rank_zero:
            results = {
                'images': images.detach().cpu().numpy(),
                'pred': {"label": pred, "conf_score": conf_score}
            }
            return results

    def test_step(self, test_model, batch):
        test_model.eval()
        indices, images, _ = batch
        images = images.to(self.devices)

        out = test_model(images)
        pred, conf_score = self.postprocessor(out, k=1)

        indices = indices.numpy()
        if self.conf.distributed:
            gathered_pred = [None for _ in range(torch.distributed.get_world_size())]

            # Remove dummy samples, they only come in distributed environment
            pred = pred[indices != -1]
            torch.distributed.gather_object(pred, gathered_pred if torch.distributed.get_rank() == 0 else None, dst=0)
            torch.distributed.barrier()
            if torch.distributed.get_rank() == 0:
                gathered_pred = np.concatenate(gathered_pred, axis=0)
                pred = gathered_pred

        if self.single_gpu_or_rank_zero:
            results = {
                'images': images.detach().cpu().numpy(),
                'pred': {"label": pred, "conf_score": conf_score}
            }
            return results

    def get_metric_with_all_outputs(self, outputs, phase: Literal['train', 'valid'], metric_factory):
        pass

    def _convert_result(self, result, class_map):
        assert "pred" in result and "images" in result
        return_preds = []
        for idx in range(len(result['images'])):
            image = result['images'][idx:idx+1]
            height, width = image.shape[-2:]
            label = result['pred']['label'][idx]
            conf_score = result['pred']['conf_score'][idx]
            return_preds.append(
                {
                    "class": int(label[0]),
                    "name": class_map[int(label[0])],
                    "conf_score": float(conf_score[0]),
                    "shape": {
                        "width": width,
                        "height": height
                    }
                }
            )
        return return_preds

    def get_predictions(self, results, class_map):
        predictions = []
        if isinstance(results, list):
            for minibatch in results:
                predictions.extend(self._convert_result(minibatch, class_map))
        elif isinstance(results, dict):
            predictions.extend(self._convert_result(results, class_map))

        return predictions



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import List

from omegaconf import DictConfig
import torch
import torch.nn as nn

from ...op.custom import SPPELAN, ELAN, AConv, ADown
from ...utils import BackboneOutput


class YOLOv9FPN(nn.Module):
    """
        YOLOv9 Path Aggregation Feature Pyramid Network (FPN) module.
        Based on https://github.com/WongKinYiu/YOLO/blob/main/yolo/model/module.py.
    """
    def __init__(
        self,
        intermediate_features_dim: List[int],
        params: DictConfig,
    ):
        super().__init__()
        self.in_channels = intermediate_features_dim
        repeat_num = params.repeat_num
        act_type = params.act_type
        bu_type = params.bu_type # bottom-up type
        assert bu_type.lower() in ['aconv', 'adown']
        bu_block = AConv if bu_type == 'aconv' else ADown
        spp_channels = params.spp_channels
        n4_channels = params.n4_channels
        p3_channels = params.p3_channels
        p4_channels = params.p4_channels
        p5_channels = params.p5_channels
        p3_to_p4_channels = params.p3_to_p4_channels
        p4_to_p5_channels = params.p4_to_p5_channels
        
        # self.use_aux_loss = params.use_aux_loss
        
        # Top-down pathway (upsampling)
        self.upsample = nn.Upsample(scale_factor=2, mode="nearest")
        self.spp_block = SPPELAN(
            in_channels=int(self.in_channels[2]),
            out_channels=spp_channels,
            act_type=act_type,
        )
        # if self.use_aux_loss:
        #     self.aux_spp_block = SPPELAN(
        #         in_channels=int(self.in_channels[2]),
        #         out_channels=spp_channels,
        #         act_type=act_type
        #     )
        # else:
        #     self.aux_spp_block = None
        
        # Top-down fusion blocks
        self.td_fusion_block_1 = ELAN(
            in_channels=int(self.in_channels[1] + spp_channels),
            out_channels=n4_channels,
            part_channels=n4_channels,
            n=repeat_num,
            layer_type="repncsp",
            act_type=act_type,
            use_identity=False
        )

        # if self.use_aux_loss:
        #     self.aux_td_fusion_block_1 = ELAN(
        #         in_channels=int(self.in_channels[1] + spp_channels),
        #         out_channels=n4_channels,
        #         part_channels=n4_channels,
        #         n=repeat_num,
        #         layer_type="repncsp",
        #         act_type=act_type,
        #         use_identity=False
        #     )
        # else:
        #     self.aux_td_fusion_block_1 = None
        
        self.td_fusion_block_2 = ELAN(
            in_channels=int(self.in_channels[0] + n4_channels),
            out_channels=p3_channels,
            part_channels=p3_channels,
            n=repeat_num,
            layer_type="repncsp",
            act_type=act_type,
            use_identity=False
        )

        # if self.use_aux_loss:
        #     self.aux_td_fusion_block_2 = ELAN(
        #         in_channels=int(self.in_channels[0] + n4_channels),
        #         out_channels=p3_channels,
        #         part_channels=p3_channels,
        #         n=repeat_num,
        #         layer_type="repncsp",
        #         act_type=act_type,
        #         use_identity=False
        #     )
        # else:
        #     self.aux_td_fusion_block_2 = None
        
        # Bottom-up pathway (downsampling)
        self.bu_conv_p3_to_p4 = bu_block(
            in_channels=p3_channels,
            out_channels=p3_to_p4_channels,
            act_type=act_type
        )
        
        self.bu_fusion_block_1 = ELAN(
            in_channels=int(p3_to_p4_channels + n4_channels),
            out_channels=p4_channels,
            part_channels=p4_channels,
            n=repeat_num,
            layer_type="repncsp",
            act_type=act_type,
            use_identity=False
        )
        
        self.bu_conv_p4_to_p5 = bu_block(
            in_channels=p4_channels, 
            out_channels=p4_to_p5_channels,
            act_type=act_type
        )
        
        self.bu_fusion_block_2 = ELAN(
            in_channels=int(p4_to_p5_channels + spp_channels),
            out_channels=p5_channels,
            part_channels=p5_channels,
            n=repeat_num,
            layer_type="repncsp",
            act_type=act_type,
            use_identity=False
        )
        
        self._intermediate_features_dim = (p3_channels, p4_channels, p5_channels)
        
        # Initialize BatchNorm layers
        def init_bn(M):
            for m in M.modules():
                if isinstance(m, nn.BatchNorm2d):
                    m.eps = 1e-3
                    m.momentum = 0.03
        self.apply(init_bn)

    def forward(self, inputs):
        """
        Args:
            inputs: List of backbone features [P3, P4, P5]
        Returns:
            Tuple[Tensor]: FPN features at different scales
        """
        [feat_p3, feat_p4, feat_p5] = inputs
        
        # Top-down pathway
        spp_feat = self.spp_block(feat_p5)  # P5 processing
        td_p4 = self.upsample(spp_feat)  # P5 -> P4
        td_p4_concat = torch.cat([td_p4, feat_p4], 1)
        td_p4_processed = self.td_fusion_block_1(td_p4_concat) # N4
        
        td_p3 = self.upsample(td_p4_processed)  # P4 -> P3
        td_p3_concat = torch.cat([td_p3, feat_p3], 1)
        p3_out = self.td_fusion_block_2(td_p3_concat)
        
        # Bottom-up pathway
        bu_p4 = self.bu_conv_p3_to_p4(p3_out)  # P3 -> P4
        bu_p4_concat = torch.cat([bu_p4, td_p4_processed], 1)
        p4_out = self.bu_fusion_block_1(bu_p4_concat)
        
        bu_p5 = self.bu_conv_p4_to_p5(p4_out)  # P4 -> P5
        bu_p5_concat = torch.cat([bu_p5, spp_feat], 1)
        p5_out = self.bu_fusion_block_2(bu_p5_concat)
        
        outputs = (p3_out, p4_out, p5_out)

        # if self.training and self.use_aux_loss:
        #     spp_a5 = self.aux_spp_block(feat_p5) # A5
        #     aux_td_p4 = self.upsample(spp_a5)
        #     aux_td_p4_concat = torch.cat([aux_td_p4, feat_p4], 1)
        #     td_a4 = self.aux_td_fusion_block_1(aux_td_p4_concat) # A4
        #     aux_td_p3 = self.upsample(td_a4)
        #     aux_td_p3_concat = torch.cat([aux_td_p3, feat_p3], 1)
        #     td_a3 = self.aux_td_fusion_block_2(aux_td_p3_concat) # A3
        #     aux_outputs = (td_a3, td_a4, spp_a5)
        #     outputs = {"outputs": outputs, "aux_outputs": aux_outputs}
        # else:
        #     aux_outputs = None
        return BackboneOutput(intermediate_features=outputs)
    
    @property
    def intermediate_features_dim(self):
        return self._intermediate_features_dim

def yolov9fpn(intermediate_features_dim, conf_model_neck, **kwargs):
    return YOLOv9FPN(intermediate_features_dim=intermediate_features_dim, params=conf_model_neck.params)


# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import List

from omegaconf import ListConfig
import torch
import torch.nn as nn
from torch import Tensor


class AnchorGenerator(nn.Module):
    """
    Module that generates anchors for a set of feature maps and
    image sizes.

    The module support computing anchors at multiple sizes and aspect ratios
    per feature map. This module assumes aspect ratio = height / width for
    each anchor.

    sizes and aspect_ratios should have the same number of elements, and it should
    correspond to the number of feature maps.

    sizes[i] and aspect_ratios[i] can have an arbitrary number of elements,
    and AnchorGenerator will output a set of sizes[i] * aspect_ratios[i] anchors
    per spatial location for feature map i.

    Args:
        sizes (Tuple[Tuple[int]]):
        aspect_ratios (Tuple[Tuple[float]]):
    """

    __annotations__ = {
        "cell_anchors": List[torch.Tensor],
    }

    def __init__(
        self,
        sizes=((128, 256, 512),),
        aspect_ratios=((0.5, 1.0, 2.0),),
        image_size=(512, 512),
    ):
        super().__init__()

        if not isinstance(sizes[0], (list, tuple, ListConfig)):
            # TODO change this
            sizes = tuple((s,) for s in sizes)
        if not isinstance(aspect_ratios[0], (list, tuple, ListConfig)):
            aspect_ratios = (aspect_ratios,) * len(sizes)

        self.sizes = sizes
        self.aspect_ratios = aspect_ratios
        self.cell_anchors = [
            self.generate_anchors(size, aspect_ratio) for size, aspect_ratio in zip(sizes, aspect_ratios)
        ]
        
        self.image_size = image_size
        self.anchors_over_all_feature_maps = None

    # TODO: https://github.com/pytorch/pytorch/issues/26792
    # For every (aspect_ratios, scales) combination, output a zero-centered anchor with those values.
    # (scales, aspect_ratios) are usually an element of zip(self.scales, self.aspect_ratios)
    # This method assumes aspect ratio = height / width for an anchor.
    def generate_anchors(
        self,
        scales: List[int],
        aspect_ratios: List[float],
        dtype: torch.dtype = torch.float32,
        device: torch.device = torch.device("cpu"),
    ) -> Tensor:
        scales = torch.as_tensor(scales, dtype=dtype, device=device)
        aspect_ratios = torch.as_tensor(aspect_ratios, dtype=dtype, device=device)
        h_ratios = torch.sqrt(aspect_ratios)
        w_ratios = 1 / h_ratios

        ws = (w_ratios[:, None] * scales[None, :]).view(-1)
        hs = (h_ratios[:, None] * scales[None, :]).view(-1)

        base_anchors = torch.stack([-ws, -hs, ws, hs], dim=1) / 2
        return base_anchors.round()

    def set_cell_anchors(self, dtype: torch.dtype, device: torch.device):
        return [cell_anchor.to(dtype).to(device) for cell_anchor in self.cell_anchors]

    def num_anchors_per_location(self) -> List[int]:
        return [len(s) * len(a) for s, a in zip(self.sizes, self.aspect_ratios)]

    # For every combination of (a, (g, s), i) in (self.cell_anchors, zip(grid_sizes, strides), 0:2),
    # output g[i] anchors that are s[i] distance apart in direction i, with the same dimensions as a.
    def grid_anchors(self, cell_anchors, grid_templates, grid_sizes: List[List[int]], strides: List[List[Tensor]]) -> List[Tensor]:
        anchors = []
        cell_anchors = cell_anchors
        torch._assert(cell_anchors is not None, "cell_anchors should not be None")
        torch._assert(
            len(grid_sizes) == len(strides) == len(cell_anchors),
            "Anchors should be Tuple[Tuple[int]] because each feature "
            "map could potentially have different sizes and aspect ratios. "
            "There needs to be a match between the number of "
            "feature maps passed and the number of sizes / aspect ratios specified.",
        )

        for grid_template, size, stride, base_anchors in zip(grid_templates, grid_sizes, strides, cell_anchors):
            grid_height_template, grid_width_template = grid_template
            grid_height, grid_width = size
            stride_height, stride_width = stride
            device = base_anchors.device

            # For output anchor, compute [x_center, y_center, x_center, y_center]
            shifts_x = torch.arange(0, grid_width, dtype=torch.int32).to(device) * stride_width
            shifts_y = torch.arange(0, grid_height, dtype=torch.int32).to(device) * stride_height
            shift_y, shift_x = torch.meshgrid(shifts_y, shifts_x, indexing="ij")
            shift_x = shift_x.reshape(-1)
            shift_y = shift_y.reshape(-1)
            shifts = torch.stack((shift_x, shift_y, shift_x, shift_y), dim=1)

            # For every (base anchor, output anchor) pair,
            # offset each zero-centered base anchor by the center of the output anchor.
            anchors.append((shifts.view(-1, 1, 4) + base_anchors.view(1, -1, 4)).reshape(-1, 4))

        return anchors

    def forward(self, feature_maps: List[Tensor]) -> List[Tensor]:
        # TODO: Use constant anchor for fx transform
        # This forces inference image size same with training phase.
        if self.anchors_over_all_feature_maps:
            return self.anchors_over_all_feature_maps

        grid_sizes = [feature_map.shape[-2:] for feature_map in feature_maps]
        # each feature_map has (b, c, h, w) shape
        grid_templates = [(feature_map[0, 0, :, 0], feature_map[0, 0, 0, :]) for feature_map in feature_maps]
        dtype, device = feature_maps[0].dtype, feature_maps[0].device
        strides = [
            [
                torch.empty(1, dtype=torch.int64).fill_(self.image_size[0] // g[0]).to(device),
                torch.empty(1, dtype=torch.int64).fill_(self.image_size[1] // g[1]).to(device),
            ]
            for g in grid_sizes
        ]
        cell_anchors = self.set_cell_anchors(dtype, device)
        anchors_over_all_feature_maps = self.grid_anchors(cell_anchors, grid_templates, grid_sizes, strides)

        self.anchors_over_all_feature_maps = anchors_over_all_feature_maps
        return anchors_over_all_feature_maps



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import Any, Optional

import torch
import torch.nn.functional as F

from ..models.utils import ModelOutput


class SegmentationPostprocessor:
    def __init__(self, conf_model):
        pass
    def __call__(self, outputs: ModelOutput, original_shape):
        pred = outputs['pred']
        H, W = original_shape[-2:]
        # upsample logits to the images' original size
        pred = F.interpolate(
            pred, size=(H, W), mode="bilinear", align_corners=False
        )
        pred = torch.max(pred, dim=1)[1]  # argmax
        return pred.detach().cpu().numpy()



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import warnings
from bisect import bisect_right
from collections import Counter

from torch.optim.lr_scheduler import _LRScheduler


class MultiStepLR(_LRScheduler):
    """Decays the learning rate of each parameter group by gamma once the
    number of epochs reaches one of the milestones.
    Args:
        optimizer (Optimizer): Wrapped optimizer.
        milestones (list): List of epoch indices. Must be increasing.
        gamma (float): Multiplicative factor of learning rate decay.
            Default: 0.1.
    Example:
        >>> # Assuming optimizer uses lr = 0.05 for all groups
        >>> # lr = 0.05     if epoch < 150
        >>> # lr = 0.005    if 150 <= epoch < 250
        >>> # lr = 0.0005   if epoch >= 250
        >>> for epoch in range(300):
        >>>     train(...)
        >>>     validate(...)
        >>>     scheduler.step()
    """
    def __init__(
        self,
        optimizer,
        scheduler_conf,
        training_epochs,
    ):
        self.milestones = Counter(scheduler_conf.milestones)
        self.gamma = scheduler_conf.gamma
        super().__init__(optimizer)

    def get_lr(self):
        if not self._get_lr_called_within_step:
            warnings.warn("To get the last learning rate computed by the scheduler, "
                          "please use `get_last_lr()`.", UserWarning, stacklevel=2)

        if self.last_epoch not in self.milestones:
            return [group['lr'] for group in self.optimizer.param_groups]
        return [
            group["lr"] * self.gamma ** self.milestones[self.last_epoch]
            for group in self.optimizer.param_groups
        ]

    def _get_closed_form_lr(self):
        milestones = sorted(self.milestones.elements())
        return [
            base_lr * self.gamma ** bisect_right(milestones, self.last_epoch)
            for base_lr in self.base_lrs
        ]



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import Dict

import torch
import torch.nn as nn
import torch.nn.functional as F

__all__ = ['PIDNetCrossEntropy', 'BoundaryLoss']

IGNORE_INDEX_NONE_VALUE = -100

NUM_OUTPUTS = 2
BALANCE_WEIGHTS = [0.4, 1.0]

OHEMTHRES = 0.9
OHEMKEEP = 131072


class PIDNetCrossEntropy(nn.Module):
    def __init__(self, ignore_index=IGNORE_INDEX_NONE_VALUE, weight=None):
        super(PIDNetCrossEntropy, self).__init__()
        self.ignore_index = ignore_index
        self.loss_fn = nn.CrossEntropyLoss(
            weight=weight,
            ignore_index=ignore_index
        )
        self.boundary_aware = False

    def _forward(self, out: torch.Tensor, target: torch.Tensor):

        return self.loss_fn(out, target)

    def forward(self, out: Dict, target: Dict):
        target = target['target']

        if self.boundary_aware:
            pred, extra_d = out['pred'], out['extra_d']
            filler = torch.ones_like(target) * self.ignore_index
            bd_label = torch.where(torch.sigmoid(extra_d[:, 0, :, :]) > 0.8, target, filler)
            return self._forward(pred, bd_label)

        pred, extra_p = out['pred'], out['extra_p']
        score = [extra_p, pred]
        return sum([w * self._forward(x, target) for (w, x) in zip(BALANCE_WEIGHTS, score)])

class PIDNetBoundaryAwareCrossEntropy(PIDNetCrossEntropy):
    def __init__(self, ignore_index=IGNORE_INDEX_NONE_VALUE, weight=None):
        super().__init__(ignore_index, weight)
        self.boundary_aware = True

# class OhemCrossEntropy(nn.Module):
#     def __init__(self, ignore_label=-1, thres=0.7, min_kept=100000, weight=None):
#         super(OhemCrossEntropy, self).__init__()
#         self.thresh = thres
#         self.min_kept = max(1, min_kept)
#         self.ignore_label = ignore_label
#         self.criterion = nn.CrossEntropyLoss(
#             weight=weight,
#             ignore_index=ignore_label,
#             reduction='none'
#         )

#     def _ce_forward(self, score, target):

#         loss = self.criterion(score, target)

#         return loss

#     def _ohem_forward(self, score, target, **kwargs):

#         pred = F.softmax(score, dim=1)
#         pixel_losses = self.criterion(score, target).contiguous().view(-1)
#         mask = target.contiguous().view(-1) != self.ignore_label

#         tmp_target = target.clone()
#         tmp_target[tmp_target == self.ignore_label] = 0
#         pred = pred.gather(1, tmp_target.unsqueeze(1))
#         pred, ind = pred.contiguous().view(-1,)[mask].contiguous().sort()
#         min_value = pred[min(self.min_kept, pred.numel() - 1)]
#         threshold = max(min_value, self.thresh)

#         pixel_losses = pixel_losses[mask][ind]
#         pixel_losses = pixel_losses[pred < threshold]
#         return pixel_losses.mean()

#     def forward(self, score, target):

#         if not (isinstance(score, list) or isinstance(score, tuple)):
#             score = [score]

#         balance_weights = BALANCE_WEIGHTS
#         sb_weights = SB_WEIGHTS
#         if len(balance_weights) == len(score):
#             functions = [self._ce_forward] * \
#                 (len(balance_weights) - 1) + [self._ohem_forward]
#             return sum([
#                 w * func(x, target)
#                 for (w, x, func) in zip(balance_weights, score, functions)
#             ])

#         elif len(score) == 1:
#             return sb_weights * self._ohem_forward(score[0], target)

#         else:
#             raise ValueError("lengths of prediction and target are not identical!")


class BoundaryLoss(nn.Module):
    def __init__(self):
        super(BoundaryLoss, self).__init__()

    @staticmethod
    def weighted_bce(bd_pre, target):
        n, c, h, w = bd_pre.size()
        log_p = bd_pre.permute(0, 2, 3, 1).contiguous().view(1, -1)
        target_t = target.view(1, -1)

        pos_index = (target_t == 1)
        neg_index = (target_t == 0)

        weight = torch.zeros_like(log_p)
        pos_num = pos_index.sum()
        neg_num = neg_index.sum()
        sum_num = pos_num + neg_num
        weight[pos_index] = neg_num * 1.0 / sum_num
        weight[neg_index] = pos_num * 1.0 / sum_num

        loss = F.binary_cross_entropy_with_logits(log_p, target_t, weight, reduction='mean')

        return loss

    def forward(self, out: Dict, target: Dict) -> torch.Tensor:
        bd_gt = target['bd_gt']
        extra_d = out['extra_d']
        return self.weighted_bce(extra_d, bd_gt)


class PIDNetLoss(nn.Module):
    def __init__(self, ignore_index=IGNORE_INDEX_NONE_VALUE, weight=None, **kwargs):
        super().__init__()

        self.cross_entropy_loss = PIDNetCrossEntropy(ignore_index=ignore_index)
        self.boundary_loss = BoundaryLoss()
        self.cross_entropy_with_boundary = PIDNetBoundaryAwareCrossEntropy(ignore_index=ignore_index)

    def forward(self, out: Dict, target: Dict):
        if out['extra_p'] is None:
            return torch.sum(out['pred'] - out['pred']) # return 0

        out['pred'] = F.interpolate(out['pred'], size=target['target'].shape[-2:], mode='bilinear', align_corners=True)
        out['extra_p'] = F.interpolate(out['extra_p'], size=target['target'].shape[-2:], mode='bilinear', align_corners=True)
        out['extra_d'] = F.interpolate(out['extra_d'], size=target['target'].shape[-2:], mode='bilinear', align_corners=True)

        cross_entropy_loss = self.cross_entropy_loss(out, target)
        boundary_loss = self.boundary_loss(out, target)
        cross_entropy_loss_with_boundary = self.cross_entropy_with_boundary(out, target)

        # TODO: return as dict
        loss = cross_entropy_loss + 20 * boundary_loss + cross_entropy_loss_with_boundary
        return loss



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import json
from dataclasses import asdict
from pathlib import Path
from statistics import mean
from typing import Dict, List, Literal, Optional, final

import torch
import torch.nn as nn
from loguru import logger
from omegaconf import DictConfig
from torch.utils.data import DataLoader
from tqdm import tqdm

from ..loggers.base import TrainingLogger
from ..utils.record import InferenceSummary, PredictionSummary, Timer
from ..utils.stats import get_params_and_flops
from .base import BasePipeline
from .task_processors.base import BaseTaskProcessor

NUM_SAMPLES = 16


class InferencePipeline(BasePipeline):
    def __init__(
        self,
        conf: DictConfig,
        task: str,
        task_processor: BaseTaskProcessor,
        model_name: str,
        model: nn.Module,
        logger: Optional[TrainingLogger],
        timer: Timer,
        test_dataloader: DataLoader,
        single_gpu_or_rank_zero: bool,
    ):
        super(InferencePipeline, self).__init__(conf, task, task_processor, model_name, model, logger, timer)
        self.test_dataloader = test_dataloader
        self.single_gpu_or_rank_zero = single_gpu_or_rank_zero

    @final
    def _is_ready(self):
        assert self.model is not None, "`self.model` is not defined!"
        return True

    @torch.no_grad()
    def inference(self):
        self._is_ready()
        self.timer.start_record(name='inference')

        num_returning_samples = 0
        returning_samples = []
        outputs = []
        for _idx, batch in enumerate(tqdm(self.test_dataloader, leave=False)):
            out = self.task_processor.test_step(self.model, batch)
            if out is not None:
                outputs.append(out)
                if num_returning_samples < NUM_SAMPLES: # TODO: Save all output or set by config
                    returning_samples.append(out)
                    num_returning_samples += len(out['pred'])

        self.timer.end_record(name='inference')
        if self.single_gpu_or_rank_zero:
            time_for_inference = self.timer.get(name='inference', as_pop=False)
            self.log_end_inference(time_for_inference=time_for_inference, valid_samples=returning_samples)

    def log_end_inference(
        self,
        time_for_inference: float,
        valid_samples: Optional[List] = None,
    ):
        self.log_results(
            prefix='inference',
            samples=valid_samples,
            elapsed_time=time_for_inference,
        )
        predictions = self.task_processor.get_predictions(valid_samples, self.logger.class_map)
        self.save_summary(time_for_inference, predictions)

    def save_summary(self, time_for_inference, predictions):
        flops, params = get_params_and_flops(self.model, self.sample_input.float())
        inference_summary = InferenceSummary(
            flops=flops,
            params=params,
            total_inference_time=time_for_inference,
            success=True,
        )

        predictions_summary = PredictionSummary(
            predictions=predictions,
            misc=None
        )

        logger.info(f"[Model stats] | Sample input: {tuple(self.sample_input.shape)} | Params: {(params/1e6):.2f}M | FLOPs: {(flops/1e9):.2f}G")
        logging_dir = self.logger.result_dir
        summary_path = Path(logging_dir) / "inference_summary.json"
        prediction_path = Path(logging_dir) / "predictions.json"

        with open(summary_path, 'w') as f:
            json.dump(asdict(inference_summary), f, indent=4)
        logger.info(f"Model inference summary saved at {str(summary_path)}")

        with open(prediction_path, 'w') as f:
            json.dump(asdict(predictions_summary), f, indent=4)
        logger.info(f"Model predictions are saved at {str(prediction_path)}")



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import json
from functools import partial
from itertools import chain
from multiprocessing.pool import ThreadPool
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import PIL.Image as Image
import torch.distributed as dist
from loguru import logger
from omegaconf import ListConfig

from .base import BaseCustomDataset, BaseSampleLoader
from .utils.constants import IMG_EXTENSIONS
from .utils.misc import natural_key


class PoseEstimationSampleLoader(BaseSampleLoader):
    def __init__(self, conf_data, train_valid_split_ratio):
        super(PoseEstimationSampleLoader, self).__init__(conf_data, train_valid_split_ratio)

    def load_data(self, split='train'):
        assert split in ['train', 'valid', 'test'], f"split should be either {['train', 'valid', 'test']}."
        data_root = Path(self.conf_data.path.root)
        split_dir = self.conf_data.path[split]
        image_dir: Path = data_root / split_dir.image
        annotation_dir: Optional[Path] = data_root / split_dir.label if split_dir.label is not None else None
        images: List[str] = []
        labels: List[str] = []
        images_and_targets: List[Dict[str, str]] = []
        if annotation_dir is not None:
            for ext in IMG_EXTENSIONS:
                for file in chain(image_dir.glob(f'*{ext}'), image_dir.glob(f'*{ext.upper()}')):
                    ann_path_maybe = annotation_dir / file.with_suffix('.txt').name
                    if not ann_path_maybe.exists():
                        continue
                    images.append(str(file))
                    labels.append(str(ann_path_maybe))
                # TODO: get paired data from regex pattern matching (self.conf_data.path.pattern)

            images = sorted(images, key=lambda k: natural_key(k))
            labels = sorted(labels, key=lambda k: natural_key(k))
            images_and_targets.extend([{'image': str(image), 'label': str(label)} for image, label in zip(images, labels)])

        else:
            for ext in IMG_EXTENSIONS:
                images_and_targets.extend([{'image': str(file), 'label': None}
                                        for file in chain(image_dir.glob(f'*{ext}'), image_dir.glob(f'*{ext.upper()}'))])
            images_and_targets = sorted(images_and_targets, key=lambda k: natural_key(k['image']))

        return images_and_targets

    def load_id_mapping(self):
        root_path = Path(self.conf_data.path.root)

        if isinstance(self.conf_data.id_mapping, ListConfig):
            return list(self.conf_data.id_mapping)
        elif isinstance(self.conf_data.id_mapping, str):
            id_mapping_path = root_path / self.conf_data.id_mapping
            with open(id_mapping_path, 'r') as f:
                id_mapping = json.load(f)
            return id_mapping
        else:
            raise ValueError(f"Invalid id_mapping type: {type(self.conf_data.id_mapping)}")

    def load_class_map(self, id_mapping):
        idx_to_class: Dict[int, str] = dict(enumerate(id_mapping))
        return {'idx_to_class': idx_to_class}

    def load_huggingface_samples(self):
        raise NotImplementedError


class PoseEstimationCustomDataset(BaseCustomDataset):

    def __init__(self, conf_data, conf_augmentation, model_name, idx_to_class,
                 split, samples, transform=None, **kwargs):
        super(PoseEstimationCustomDataset, self).__init__(
            conf_data, conf_augmentation, model_name, idx_to_class,
            split, samples, transform, **kwargs
        )
        flattened_samples = []
        # label field must be filled
        for sample in self.samples:
            flattened_sample = {}
            with open(sample['label'], 'r') as f:
                lines = f.readlines()
                f.close()
            flattened_sample = [{'image': sample['image'], 'label': line.strip()} for line in lines]
            flattened_samples += flattened_sample
        self.samples = flattened_samples

        # Build flip map. This is needed when try randomflip augmentation.
        if split == 'train':
            trasnform_names = {transform_conf['name'] for transform_conf in conf_augmentation[split]}
            flips = {'randomhorizontalflip', 'randomverticalflip'}
            if len(trasnform_names.intersection(flips)) > 0:
                class_to_idx = {self._idx_to_class[i]['name']: i for i in self._idx_to_class}
                self.flip_indices = np.zeros(self._num_classes).astype('int')
                for idx in self._idx_to_class:
                    idx_swap = self._idx_to_class[idx]['swap']
                    assert idx_swap is not None, "To apply flip transform, keypoint swap info must be filled."
                    self.flip_indices[idx] = class_to_idx[idx_swap] if idx_swap else -1

    def cache_dataset(self):
        logger.info(f'Caching | Loading samples of {self.mode} to memory... This can take minutes.')

        def _load(i, samples):
            image = Image.open(Path(samples[i]['image'])).convert('RGB')
            return i, image

        num_threads = 8 # TODO: Compute appropriate num_threads
        all_indices = list(range(len(self.samples)))
        load_imgs = ThreadPool(num_threads).imap(
            partial(_load, samples=self.samples),
            all_indices
        )
        for i, image in load_imgs:
            self.samples[i]['image'] = image

        self.cache = True

    def __getitem__(self, index):
        img = self.samples[index]['image']
        ann = self.samples[index]['label'] # TODO: Pose estimation is not assuming that label can be None now

        if not self.cache:
            img = Image.open(Path(img)).convert('RGB')

        w, h = img.size

        outputs = {}
        outputs.update({'indices': index})
        if ann is None:
            out = self.transform(image=img)
            outputs.update({'pixel_values': out['image'], 'org_shape': (h, w)})
            return outputs

        ann = ann.split(' ')
        bbox = ann[-4:]
        keypoints = ann[:-4]

        bbox = np.array(bbox).astype('float32')[np.newaxis, ...]
        keypoints = np.array(keypoints).reshape(-1, 3).astype('float32')[np.newaxis, ...]

        out = self.transform(image=img, bbox=bbox, keypoint=keypoints, dataset=self)

        # Use only one instance keypoints
        outputs.update({'pixel_values': out['image'], 'keypoints': out['keypoint'][0]})
        if self._split in ['train', 'training']:
            return outputs

        assert self._split in ['val', 'valid', 'test']
        # outputs.update({'org_img': org_img, 'org_shape': (h, w)})  # TODO: return org_img with batch_size > 1
        outputs.update({'org_shape': (h, w)})
        return outputs



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import List

import numpy as np

from ..base import BaseMetric


class PoseEstimationMetricAdaptor:
    '''
        Adapter to process redundant operations for the metrics.
    '''
    def __init__(self, metric_names) -> None:
        self.metric_names = metric_names

    def __call__(self, predictions: List[dict], targets: List[dict]):
        return {} # Do nothing


class PCK(BaseMetric):
    def __init__(self, num_classes, classwise_analysis, **kwargs):
        # TODO: Select metrics by user
        metric_name = 'PCK'
        if classwise_analysis: # TODO: Implement classwise analysis
            raise NotImplementedError('Classwise analysis is not supported for PCK metric')
        super().__init__(metric_name=metric_name, num_classes=num_classes, classwise_analysis=classwise_analysis)
        # TODO: Get from config
        self.thr = 0.05
        self.input_size = (256, 256)

    def _calc_distances(self, preds, gts, mask, norm_factor):
        N, K, _ = preds.shape
        # set mask=0 when norm_factor==0
        _mask = mask.copy()
        _mask[np.where((norm_factor == 0).sum(1))[0], :] = False

        distances = np.full((N, K), -1, dtype=np.float32)
        # handle invalid values
        norm_factor[np.where(norm_factor <= 0)] = 1e6
        distances[_mask] = np.linalg.norm(((preds - gts) / norm_factor[:, None, :])[_mask], axis=-1)
        return distances.T

    def _distance_acc(self, distances, thr):
        distance_valid = distances != -1
        num_distance_valid = distance_valid.sum()
        if num_distance_valid > 0:
            return (distances[distance_valid] < thr).sum() / num_distance_valid
        return -1

    def keypoint_pck_accuracy(self, pred, gt, mask, thr, norm_factor):
        distances = self._calc_distances(pred, gt, mask, norm_factor)
        acc = np.array([self._distance_acc(d, thr) for d in distances])
        valid_acc = acc[acc >= 0]
        cnt = len(valid_acc)
        avg_acc = valid_acc.mean() if cnt > 0 else 0.0
        # ``acc`` contains accuracy of each keypoint
        return acc, avg_acc, cnt

    def calibrate(self, pred, target, **kwargs):
        N, _, _ = pred.shape

        mask = target[..., 2] > 0
        target = target[..., :2]

        # if normalize is None:
        #     normalize = np.tile(np.array([[H, W]]), (N, 1))
        # Normalize by box
        normalize = np.tile(np.array([[self.input_size[0], self.input_size[1]]]), (N, 1))

        acc, avg_acc, cnt = self.keypoint_pck_accuracy(pred, target, mask, self.thr, normalize)
        self.metric_meter.update(avg_acc)



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

"""
Based on the RT-DETR implementation of lyuwenyu.
https://github.com/lyuwenyu/RT-DETR/blob/main/rtdetr_pytorch/src/zoo/rtdetr/hybrid_encoder.py
"""
import copy
from typing import List

from omegaconf import DictConfig
import torch 
import torch.nn as nn 
import torch.nn.functional as F 

from ...utils import BackboneOutput
from ...op.registry import ACTIVATION_REGISTRY
from ...op.custom import RepVGGBlock, ConvLayer, CSPRepLayer


# transformer
class TransformerEncoderLayer(nn.Module):
    def __init__(self,
                 d_model,
                 nhead,
                 dim_feedforward=2048,
                 dropout=0.1,
                 activation="relu",
                 normalize_before=False):
        super().__init__()
        self.normalize_before = normalize_before

        self.self_attn = nn.MultiheadAttention(d_model, nhead, dropout, batch_first=True)

        self.linear1 = nn.Linear(d_model, dim_feedforward)
        self.dropout = nn.Dropout(dropout)
        self.linear2 = nn.Linear(dim_feedforward, d_model)

        self.norm1 = nn.LayerNorm(d_model)
        self.norm2 = nn.LayerNorm(d_model)
        self.dropout1 = nn.Dropout(dropout)
        self.dropout2 = nn.Dropout(dropout)
        self.activation = ACTIVATION_REGISTRY[activation]()

    @staticmethod
    def with_pos_embed(tensor, pos_embed):
        return tensor if pos_embed is None else tensor + pos_embed

    def forward(self, src, src_mask=None, pos_embed=None) -> torch.Tensor:
        residual = src
        if self.normalize_before:
            src = self.norm1(src)
        q = k = self.with_pos_embed(src, pos_embed)
        src, _ = self.self_attn(q, k, value=src, attn_mask=src_mask)

        src = residual + self.dropout1(src)
        if not self.normalize_before:
            src = self.norm1(src)

        residual = src
        if self.normalize_before:
            src = self.norm2(src)
        src = self.linear2(self.dropout(self.activation(self.linear1(src))))
        src = residual + self.dropout2(src)
        if not self.normalize_before:
            src = self.norm2(src)
        return src


class TransformerEncoder(nn.Module):
    def __init__(self, encoder_layer, num_layers, norm=None):
        super(TransformerEncoder, self).__init__()
        self.layers = nn.ModuleList([copy.deepcopy(encoder_layer) for _ in range(num_layers)])
        self.num_layers = num_layers
        self.norm = norm

    def forward(self, src, src_mask=None, pos_embed=None) -> torch.Tensor:
        output = src
        for layer in self.layers:
            output = layer(output, src_mask=src_mask, pos_embed=pos_embed)

        if self.norm is not None:
            output = self.norm(output)

        return output


class HybridEncoder(nn.Module):
    
    def __init__(
        self,
        intermediate_features_dim: List[int],
        params: DictConfig,    
    ):
        super().__init__()
        self.in_channels = intermediate_features_dim
        self.feat_strides = [8, 16, 32] # TODO: Get from other source
        # TODO: Get from config
        self.hidden_dim = params.hidden_dim
        self.use_encoder_idx = params.use_encoder_idx
        self.num_encoder_layers = params.num_encoder_layers
        self.pe_temperature = params.pe_temperature
        self.eval_spatial_size = None

        nhead = params.num_attention_heads
        dim_feedforward = params.dim_feedforward
        dropout = params.dropout
        enc_act = params.attn_act_type
        expansion = params.expansion
        depth_mult = params.depth_mult
        act = params.conv_act_type

        self.out_channels = [self.hidden_dim for _ in range(len(self.in_channels))]
        self.out_strides = self.feat_strides
        
        # channel projection
        self.input_proj = nn.ModuleList()
        for in_channel in self.in_channels:
            self.input_proj.append(
                nn.Sequential(
                    nn.Conv2d(in_channel, self.hidden_dim, kernel_size=1, bias=False),
                    nn.BatchNorm2d(self.hidden_dim)
                )
            )

        # encoder transformer
        encoder_layer = TransformerEncoderLayer(
            self.hidden_dim, 
            nhead=nhead,
            dim_feedforward=dim_feedforward, 
            dropout=dropout,
            activation=enc_act)

        self.encoder = nn.ModuleList([
            TransformerEncoder(copy.deepcopy(encoder_layer), self.num_encoder_layers) for _ in range(len(self.use_encoder_idx))
        ])

        # top-down fpn
        self.lateral_convs = nn.ModuleList()
        self.fpn_blocks = nn.ModuleList()
        for _ in range(len(self.in_channels) - 1, 0, -1):
            self.lateral_convs.append(ConvLayer(self.hidden_dim, self.hidden_dim, kernel_size=1, stride=1, act_type=act))
            self.fpn_blocks.append(
                CSPRepLayer(self.hidden_dim * 2, self.hidden_dim, round(3 * depth_mult), act=act, expansion=expansion, use_identity=False)
            )

        # bottom-up pan
        self.downsample_convs = nn.ModuleList()
        self.pan_blocks = nn.ModuleList()
        for _ in range(len(self.in_channels) - 1):
            self.downsample_convs.append(
                ConvLayer(self.hidden_dim, self.hidden_dim, kernel_size=3, stride=2, act_type=act)
            )
            self.pan_blocks.append(
                CSPRepLayer(self.hidden_dim * 2, self.hidden_dim, round(3 * depth_mult), act=act, expansion=expansion, use_identity=False)
            )
        self._reset_parameters()

    @property
    def intermediate_features_dim(self):
        return self.out_channels

    def _reset_parameters(self):
        if self.eval_spatial_size:
            for idx in self.use_encoder_idx:
                stride = self.feat_strides[idx]
                pos_embed = self.build_2d_sincos_position_embedding(
                    self.eval_spatial_size[1] // stride, self.eval_spatial_size[0] // stride,
                    self.hidden_dim, self.pe_temperature)
                setattr(self, f'pos_embed{idx}', pos_embed)
                # self.register_buffer(f'pos_embed{idx}', pos_embed)

    @staticmethod
    def build_2d_sincos_position_embedding(w, h, embed_dim=256, temperature=10000.):
        '''
        '''
        grid_w = torch.arange(int(w), dtype=torch.float32)
        grid_h = torch.arange(int(h), dtype=torch.float32)
        grid_w, grid_h = torch.meshgrid(grid_w, grid_h, indexing='ij')
        assert embed_dim % 4 == 0, \
            'Embed dimension must be divisible by 4 for 2D sin-cos position embedding'
        pos_dim = embed_dim // 4
        omega = torch.arange(pos_dim, dtype=torch.float32) / pos_dim
        omega = 1. / (temperature ** omega)

        out_w = grid_w.flatten()[..., None] @ omega[None]
        out_h = grid_h.flatten()[..., None] @ omega[None]

        return torch.concat([out_w.sin(), out_w.cos(), out_h.sin(), out_h.cos()], dim=1)[None, :, :]

    def forward(self, feats):
        assert len(feats) == len(self.in_channels)
        proj_feats = [self.input_proj[i](feat) for i, feat in enumerate(feats)]
        
        # encoder
        if self.num_encoder_layers > 0:
            for i, enc_ind in enumerate(self.use_encoder_idx):
                h, w = proj_feats[enc_ind].shape[2:]
                # flatten [B, C, H, W] to [B, HxW, C]
                src_flatten = proj_feats[enc_ind].flatten(2).permute(0, 2, 1)
                if self.training or self.eval_spatial_size is None:
                    pos_embed = self.build_2d_sincos_position_embedding(
                        w, h, self.hidden_dim, self.pe_temperature).to(src_flatten.device)
                else:
                    pos_embed = getattr(self, f'pos_embed{enc_ind}', None).to(src_flatten.device)

                memory = self.encoder[i](src_flatten, pos_embed=pos_embed)
                proj_feats[enc_ind] = memory.permute(0, 2, 1).reshape(-1, self.hidden_dim, h, w).contiguous()
                # print([x.is_contiguous() for x in proj_feats ])

        # broadcasting and fusion
        inner_outs = [proj_feats[-1]]
        for idx in range(len(self.in_channels) - 1, 0, -1):
            feat_high = inner_outs[0]
            feat_low = proj_feats[idx - 1]
            feat_high = self.lateral_convs[len(self.in_channels) - 1 - idx](feat_high)
            inner_outs[0] = feat_high
            upsample_feat = F.interpolate(feat_high, scale_factor=2., mode='nearest')
            inner_out = self.fpn_blocks[len(self.in_channels)-1-idx](torch.concat([upsample_feat, feat_low], dim=1))
            inner_outs.insert(0, inner_out)

        outs = [inner_outs[0]]
        for idx in range(len(self.in_channels) - 1):
            feat_low = outs[-1]
            feat_high = inner_outs[idx + 1]
            downsample_feat = self.downsample_convs[idx](feat_low)
            out = self.pan_blocks[idx](torch.concat([downsample_feat, feat_high], dim=1))
            outs.append(out)

        return BackboneOutput(intermediate_features=outs)

def rtdetr_hybrid_encoder(intermediate_features_dim, conf_model_neck, **kwargs):
    return HybridEncoder(intermediate_features_dim=intermediate_features_dim, params=conf_model_neck.params)



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from pathlib import Path
from typing import Dict, List, Literal, Optional, Tuple, Union

from .base import TrainingLogger


def build_logger(conf, task: str, model_name: str, step_per_epoch: int, class_map: Dict[int, str], num_sample_images: int, result_dir: Union[Path, str], epoch: Optional[int] = None):
    training_logger = TrainingLogger(conf,
                                     task=task, model=model_name,
                                     step_per_epoch=step_per_epoch,
                                     class_map=class_map, num_sample_images=num_sample_images,
                                     result_dir=result_dir,
                                     epoch=epoch)

    return training_logger



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------
"""
Based on the RT-DETR implementation of lyuwenyu.
https://github.com/lyuwenyu/RT-DETR/blob/main/rtdetr_pytorch/src/zoo/rtdetr/denoising.py
"""

import torch


def inverse_sigmoid(x: torch.Tensor, eps: float=1e-5) -> torch.Tensor:
    x = x.clip(min=0., max=1.)
    return torch.log(x.clip(min=eps) / (1 - x).clip(min=eps))

def box_cxcywh_to_xyxy(x):
    x_c, y_c, w, h = x.unbind(-1)
    b = [(x_c - 0.5 * w), (y_c - 0.5 * h),
         (x_c + 0.5 * w), (y_c + 0.5 * h)]
    return torch.stack(b, dim=-1)


def box_xyxy_to_cxcywh(x):
    x0, y0, x1, y1 = x.unbind(-1)
    b = [(x0 + x1) / 2, (y0 + y1) / 2,
         (x1 - x0), (y1 - y0)]
    return torch.stack(b, dim=-1)


def get_contrastive_denoising_training_group(targets,
                                             num_classes,
                                             num_queries,
                                             class_embed,
                                             num_denoising=100,
                                             label_noise_ratio=0.5,
                                             box_noise_scale=1.0,):
    """cnd"""
    if num_denoising <= 0:
        return None, None, None, None

    num_gts = [len(t['labels']) for t in targets]
    device = targets[0]['labels'].device

    max_gt_num = max(num_gts)
    if max_gt_num == 0:
        return None, None, None, None

    num_group = num_denoising // max_gt_num
    num_group = 1 if num_group == 0 else num_group
    # pad gt to max_num of a batch
    bs = len(num_gts)

    input_query_class = torch.full([bs, max_gt_num], num_classes, dtype=torch.int32, device=device)
    input_query_bbox = torch.zeros([bs, max_gt_num, 4], device=device)
    pad_gt_mask = torch.zeros([bs, max_gt_num], dtype=torch.bool, device=device)

    for i in range(bs):
        num_gt = num_gts[i]
        if num_gt > 0:
            boxes = box_xyxy_to_cxcywh(targets[i]['boxes'])
            input_query_class[i, :num_gt] = targets[i]['labels']
            input_query_bbox[i, :num_gt] = boxes
            pad_gt_mask[i, :num_gt] = 1
    # each group has positive and negative queries.
    input_query_class = input_query_class.tile([1, 2 * num_group])
    input_query_bbox = input_query_bbox.tile([1, 2 * num_group, 1])
    pad_gt_mask = pad_gt_mask.tile([1, 2 * num_group])
    # positive and negative mask
    negative_gt_mask = torch.zeros([bs, max_gt_num * 2, 1], device=device)
    negative_gt_mask[:, max_gt_num:] = 1
    negative_gt_mask = negative_gt_mask.tile([1, num_group, 1])
    positive_gt_mask = 1 - negative_gt_mask
    # contrastive denoising training positive index
    positive_gt_mask = positive_gt_mask.squeeze(-1) * pad_gt_mask
    dn_positive_idx = torch.nonzero(positive_gt_mask)[:, 1]
    dn_positive_idx = torch.split(dn_positive_idx, [n * num_group for n in num_gts])
    # total denoising queries
    num_denoising = int(max_gt_num * 2 * num_group)

    if label_noise_ratio > 0:
        mask = torch.rand_like(input_query_class, dtype=torch.float) < (label_noise_ratio * 0.5)
        # randomly put a new one here
        new_label = torch.randint_like(mask, 0, num_classes, dtype=input_query_class.dtype)
        input_query_class = torch.where(mask & pad_gt_mask, new_label, input_query_class)

    if box_noise_scale > 0:
        known_bbox = box_cxcywh_to_xyxy(input_query_bbox)
        diff = torch.tile(input_query_bbox[..., 2:] * 0.5, [1, 1, 2]) * box_noise_scale
        rand_sign = torch.randint_like(input_query_bbox, 0, 2) * 2.0 - 1.0
        rand_part = torch.rand_like(input_query_bbox)
        rand_part = (rand_part + 1.0) * negative_gt_mask + rand_part * (1 - negative_gt_mask)
        rand_part *= rand_sign
        known_bbox += rand_part * diff
        known_bbox.clip_(min=0.0, max=1.0)
        input_query_bbox = box_xyxy_to_cxcywh(known_bbox)
        input_query_bbox = inverse_sigmoid(input_query_bbox)

    input_query_class = class_embed(input_query_class)

    tgt_size = num_denoising + num_queries
    attn_mask = torch.full([tgt_size, tgt_size], False, dtype=torch.bool, device=device)
    # match query cannot see the reconstruction
    attn_mask[num_denoising:, :num_denoising] = True

    # reconstruct cannot see each other
    for i in range(num_group):
        if i == 0:
            attn_mask[max_gt_num * 2 * i: max_gt_num * 2 * (i + 1), max_gt_num * 2 * (i + 1): num_denoising] = True
        if i == num_group - 1:
            attn_mask[max_gt_num * 2 * i: max_gt_num * 2 * (i + 1), :max_gt_num * i * 2] = True
        else:
            attn_mask[max_gt_num * 2 * i: max_gt_num * 2 * (i + 1), max_gt_num * 2 * (i + 1): num_denoising] = True
            attn_mask[max_gt_num * 2 * i: max_gt_num * 2 * (i + 1), :max_gt_num * 2 * i] = True

    dn_meta = {
        "dn_positive_idx": dn_positive_idx,
        "dn_num_group": num_group,
        "dn_num_split": [num_denoising, num_queries]
    }

    return input_query_class, input_query_bbox, attn_mask, dn_meta



from typing import Dict, Optional

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor


class SegCrossEntropyLoss(nn.Module):
    def __init__(self, weight: Optional[Tensor]=None, size_average=None, ignore_index: int=-100,
                 reduce=None, label_smoothing: float=0.0, **kwargs):
        super(SegCrossEntropyLoss, self).__init__()
        self.loss_fn = nn.CrossEntropyLoss(weight=weight, size_average=size_average, ignore_index=ignore_index,
                                           reduce=reduce, reduction='mean', label_smoothing=label_smoothing)

    def forward(self, out: Dict, target: Dict) -> Dict:
        pred = out['pred']
        target = target['target']

        H, W = target.shape[-2:]
        # upsample logits to the images' original size
        pred = F.interpolate(
            pred, size=(H, W), mode="bilinear", align_corners=False
        )

        loss = self.loss_fn(pred, target)
        return loss



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

"""
This code is modified from https://pytorch.org/docs/stable/_modules/torch/optim/lr_scheduler.html#CosineAnnealingWarmRestarts .
"""
import math
import warnings

import torch
from loguru import logger
from torch.optim.lr_scheduler import _LRScheduler


class CosineAnnealingWarmRestartsWithCustomWarmUp(_LRScheduler):
    r"""Set the learning rate of each parameter group using a cosine annealing
    schedule, where :math:`\eta_{max}` is set to the initial lr, :math:`T_{cur}`
    is the number of epochs since the last restart and :math:`T_{i}` is the number
    of epochs between two warm restarts in SGDR:

    .. math::
        \eta_t = \eta_{min} + \frac{1}{2}(\eta_{max} - \eta_{min})\left(1 +
        \cos\left(\frac{T_{cur}}{T_{i}}\pi\right)\right)

    When :math:`T_{cur}=T_{i}`, set :math:`\eta_t = \eta_{min}`.
    When :math:`T_{cur}=0` after restart, set :math:`\eta_t=\eta_{max}`.

    It has been proposed in
    `SGDR: Stochastic Gradient Descent with Warm Restarts`_.

    Args:
        optimizer (Optimizer): Wrapped optimizer.
        warmup_iters (int): The number of steps that the scheduler finishes to warmup the learning rate. Default: 5.
        total_iters (int): Maximum number of iterations. Default: 5.
        warmup_bias_lr (int): Starting learning rate for warmup period. Default: 0.
        iters_per_phase (int): Number of iterations for the first restart. Originally named as `T_0`. Default: 0.
        T_mult (int, optional): A factor increases :math:`T_{i}` after a restart. Default: 1.
        min_lr (float, optional): Minimum learning rate. Originally named as `eta_min`. Default: 0.
        last_epoch (int, optional): The index of last epoch. Default: -1.
        verbose (bool): If ``True``, prints a message to stdout for
            each update. Default: ``False``.

    .. _SGDR\: Stochastic Gradient Descent with Warm Restarts:
        https://arxiv.org/abs/1608.03983
    """

    def __init__(
        self,
        optimizer,
        scheduler_conf,
        training_epochs,
    ):
        total_iters = training_epochs
        iters_per_phase = scheduler_conf.iters_per_phase

        if iters_per_phase <= 0 or not isinstance(iters_per_phase, int):
            T_0_maybe = total_iters // 10 if total_iters // 10 != 0 else total_iters
            assert T_0_maybe > 0 and isinstance(T_0_maybe, int)
            logger.info(f"Original T_0 is invalid {iters_per_phase}! Prefer to set T_0 as {T_0_maybe}.")
            iters_per_phase = T_0_maybe
        #if T_mult < 1 or not isinstance(T_mult, int):
        #    raise ValueError("Expected integer T_mult >= 1, but got {}".format(T_mult))
        if iters_per_phase > total_iters:
            iters_per_phase = total_iters

        self.T_0 = iters_per_phase
        self.T_i = iters_per_phase
        self.T_mult = 1 # @illian01: fix as 1 for simplicity
        self.T_i, self.remain_iters = self.get_reassigned_t_i(self.T_0, self.T_i * self.T_mult, total_iters)
        self.total_iters = total_iters
        self.eta_min = scheduler_conf.min_lr
        self.T_cur = 0 # @illian01: fix as 0 since last_epoch is set to default
        self.warmup_bias_lr = scheduler_conf.warmup_bias_lr
        self.warmup_iters = scheduler_conf.warmup_epochs
        super().__init__(optimizer)

    def get_lr(self):
        if not self._get_lr_called_within_step:
            warnings.warn("To get the last learning rate computed by the scheduler, "
                          "please use `get_last_lr()`.", UserWarning, stacklevel=2)

        if self.last_epoch >= 0 and self.last_epoch < self.warmup_iters:
            return [self.warmup_bias_lr + (float(self.last_epoch + 1) / float(max(1, self.warmup_iters))) * (base_lr - self.warmup_bias_lr)
                    for base_lr in self.base_lrs]

        return [self.eta_min + (base_lr - self.eta_min) * (1 + math.cos(math.pi * self.T_cur / self.T_i)) / 2
                for base_lr in self.base_lrs]

    @staticmethod
    def get_reassigned_t_i(current_t_i, next_t_i, remain_epochs):
        """adjust T_i to finish at last epoch in overall schedule
        """
        if remain_epochs == 0:
            return current_t_i, 0

        assert remain_epochs >= current_t_i, f"{remain_epochs}, {current_t_i}"

        if remain_epochs < current_t_i + next_t_i:
            return remain_epochs, remain_epochs

        return current_t_i, remain_epochs

    def _step_without_given_epoch(self) -> int:
        if self.last_epoch < 0:
            epoch = 0
            return epoch

        epoch = self.last_epoch + 1
        self.T_cur = self.T_cur + 1
        if self.T_cur >= self.T_i:
            self.T_cur = self.T_cur - self.T_i
            self.remain_iters = self.remain_iters - self.T_i
            self.T_i = self.T_i * self.T_mult
            self.T_i, self.remain_iters = self.get_reassigned_t_i(self.T_i, self.T_i * self.T_mult, self.remain_iters)
        return epoch

    def step(self, epoch=None):
        """Step could be called after every batch update

        Example:
            >>> # xdoctest: +SKIP("Undefined vars")
            >>> scheduler = CosineAnnealingWarmRestarts(optimizer, T_0, T_mult)
            >>> iters = len(dataloader)
            >>> for epoch in range(20):
            >>>     for i, sample in enumerate(dataloader):
            >>>         inputs, labels = sample['inputs'], sample['labels']
            >>>         optimizer.zero_grad()
            >>>         outputs = net(inputs)
            >>>         loss = criterion(outputs, labels)
            >>>         loss.backward()
            >>>         optimizer.step()
            >>>         scheduler.step(epoch + i / iters)

        This function can be called in an interleaved way.

        Example:
            >>> # xdoctest: +SKIP("Undefined vars")
            >>> scheduler = CosineAnnealingWarmRestarts(optimizer, T_0, T_mult)
            >>> for epoch in range(20):
            >>>     scheduler.step()
            >>> scheduler.step(26)
            >>> scheduler.step() # scheduler.step(27), instead of scheduler(20)
        """

        if epoch is None:
            epoch = self._step_without_given_epoch()
        else:
            if epoch < 0:
                raise ValueError("Expected non-negative epoch, but got {}".format(epoch))

            if epoch >= self.T_0:
                if self.T_mult == 1:
                    self.T_cur = epoch % self.T_0
                else:
                    n = int(math.log((epoch / self.T_0 * (self.T_mult - 1) + 1), self.T_mult))
                    self.T_cur = epoch - self.T_0 * (self.T_mult ** n - 1) / (self.T_mult - 1)
                    self.T_i = self.T_0 * self.T_mult ** (n)
                    self.T_i, _ = self.get_reassigned_t_i(self.T_i, self.T_i * self.T_mult, self.total_iters - epoch)
            else:
                self.T_cur = epoch
        self.last_epoch = math.floor(epoch)

        class _enable_get_lr_call:

            def __init__(self, o):
                self.o = o

            def __enter__(self):
                self.o._get_lr_called_within_step = True
                return self

            def __exit__(self, type, value, traceback):
                self.o._get_lr_called_within_step = False
                return self

        with _enable_get_lr_call(self):
            for i, data in enumerate(zip(self.optimizer.param_groups, self.get_lr())):
                param_group, lr = data
                param_group['lr'] = lr
                self.print_lr(self.verbose, i, lr, epoch)

        self._last_lr = [group['lr'] for group in self.optimizer.param_groups]



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import torch
import torch.fx as fx
import torch.nn as nn

__all__ = ['convert_graphmodule', 'save_graphmodule']

def _convert_graphmodule(model: nn.Module) -> fx.GraphModule:
    try:
        _graph = fx.Tracer().trace(model)
        model = fx.GraphModule(model, _graph)
        return model
    except Exception as e:
        raise e


def convert_graphmodule(model: nn.Module) -> fx.GraphModule:
    return _convert_graphmodule(model)


def save_graphmodule(model: nn.Module, f):
    model: fx.GraphModule = _convert_graphmodule(model)
    torch.save(model, f)



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import warnings

import torch
from loguru import logger
from torch.optim.lr_scheduler import _LRScheduler


class StepLR(_LRScheduler):
    """Decays the learning rate of each parameter group by gamma every
    step_size epochs. Notice that such decay can happen simultaneously with
    other changes to the learning rate from outside this scheduler. When
    last_epoch=-1, sets initial lr as lr.

    Args:
        optimizer (Optimizer): Wrapped optimizer.
        step_size (int): Period of learning rate decay.
        gamma (float): Multiplicative factor of learning rate decay.
            Default: 0.1.
        last_epoch (int): The index of last epoch. Default: -1.
        verbose (bool): If ``True``, prints a message to stdout for
            each update. Default: ``False``.

    Example:
        >>> # xdoctest: +SKIP
        >>> # Assuming optimizer uses lr = 0.05 for all groups
        >>> # lr = 0.05     if epoch < 30
        >>> # lr = 0.005    if 30 <= epoch < 60
        >>> # lr = 0.0005   if 60 <= epoch < 90
        >>> # ...
        >>> scheduler = StepLR(optimizer, step_size=30, gamma=0.1)
        >>> for epoch in range(100):
        >>>     train(...)
        >>>     validate(...)
        >>>     scheduler.step()
    """

    def __init__(
        self,
        optimizer,
        scheduler_conf,
        training_epochs,
    ):
        self.step_size = scheduler_conf.iters_per_phase
        self.gamma = scheduler_conf.gamma
        self.end_epoch = scheduler_conf.end_epoch

        super().__init__(optimizer)

    def get_lr(self):
        if not self._get_lr_called_within_step:
            warnings.warn("To get the last learning rate computed by the scheduler, "
                          "please use `get_last_lr()`.", UserWarning, stacklevel=2)

        if (self.last_epoch > self.end_epoch) or (self.last_epoch == 0) or (self.last_epoch % self.step_size != 0):
            return [group['lr'] for group in self.optimizer.param_groups]
        return [group['lr'] * self.gamma
                for group in self.optimizer.param_groups]

    def _get_closed_form_lr(self):
        return [base_lr * self.gamma ** (self.last_epoch // self.step_size)
                for base_lr in self.base_lrs]



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

import collections
import itertools
import math
from typing import Any, Dict, List, Optional, Tuple, Union

import torch
import torch.nn as nn
from torch import Tensor
from torch.fx.proxy import Proxy

from ..op.custom import LayerScale2d, Pool2d
from ..op.depth import DropPath
from ..op.registry import ACTIVATION_REGISTRY, NORM_REGISTRY
from ..utils import BackboneOutput, FXTensorType


class Pooling(nn.Module):
    """
    Implementation of pooling for PoolFormer
    """
    def __init__(self, pool_size=3):
        super().__init__()
        self.pool = Pool2d(method='avg', kernel_size=pool_size, stride=1, padding=pool_size // 2, count_include_pad=False)

    def forward(self, x):
        return self.pool(x) - x


class Image2Sequence(nn.Module):

    def __init__(self, contiguous=False):
        super().__init__()
        self.contiguous = contiguous

    def forward(self, x: Union[Tensor, Proxy]):
        x = x.flatten(2).transpose(1, 2)
        if self.contiguous:
            x = x.contiguous()
        return x

class MultiHeadAttention(nn.Module):
    def __init__(
        self,
        hidden_size,
        num_attention_heads,
        attention_hidden_size = None,
        value_hidden_size = None,
        attention_scale = None,
        attention_dropout_prob = 0.0,
        use_qkv_bias = True,
        use_attention_bias = False,
        use_cross_attention = False,
        output_with_attentions = False,
        sequence_reduction_ratio = 1,
        attention_bias_resolution = 16,
    ) -> None:
        super().__init__()

        attention_hidden_size = attention_hidden_size if attention_hidden_size is not None else hidden_size
        value_hidden_size = value_hidden_size if value_hidden_size is not None else attention_hidden_size

        if attention_hidden_size % num_attention_heads != 0:
            raise ValueError(
                f"The hidden size {attention_hidden_size,} is not a multiple of the number of attention "
                f"heads {num_attention_heads}."
            )

        if value_hidden_size % num_attention_heads != 0:
            raise ValueError(
                f"The hidden size {value_hidden_size,} is not a multiple of the number of attention "
                f"heads {num_attention_heads}."
            )

        self.num_attention_heads = num_attention_heads
        self.attention_head_size = int(attention_hidden_size / num_attention_heads)
        self.value_attention_head_size = int(value_hidden_size / num_attention_heads)

        self.head_size = self.num_attention_heads * self.attention_head_size
        self.value_head_size = self.num_attention_heads * self.value_attention_head_size
        self.attention_scale = attention_scale if attention_scale is not None \
            else math.sqrt(self.attention_head_size)


        self.query = nn.Linear(hidden_size, self.head_size, bias=use_qkv_bias)  # ... x C -> ... x C_qk
        self.key = nn.Linear(hidden_size, self.head_size, bias=use_qkv_bias)  # ... x C -> ... x C_qk
        self.value = nn.Linear(hidden_size, self.value_head_size, bias=use_qkv_bias)  # ... x C -> ... x C_v

        self.linear = nn.Linear(self.value_head_size, hidden_size)  # ... x C_v -> ... x C

        self.dropout = nn.Dropout(attention_dropout_prob)
        self.output_with_attentions = output_with_attentions

        self.use_sequence_reduction = False
        if sequence_reduction_ratio > 1:
            self.use_sequence_reduction = True
            self.sr = nn.Conv2d(
                hidden_size, hidden_size, kernel_size=sequence_reduction_ratio, stride=sequence_reduction_ratio
            )
            self.sr_layer_norm = nn.LayerNorm(hidden_size)

        self.use_attention_bias = use_attention_bias
        if self.use_attention_bias:
            # See https://github.com/snap-research/EfficientFormer/blob/main/models/efficientformer.py#L48-L61
            assert attention_bias_resolution is not None
            points = list(itertools.product(range(attention_bias_resolution), range(attention_bias_resolution)))
            len(points)
            attention_offsets = {}
            idxs = []
            for p1 in points:
                for p2 in points:
                    offset = (abs(p1[0] - p2[0]), abs(p1[1] - p2[1]))
                    if offset not in attention_offsets:
                        attention_offsets[offset] = len(attention_offsets)
                    idxs.append(attention_offsets[offset])

            self.attention_biases = torch.nn.Parameter(torch.zeros(self.num_attention_heads, 49))
            self.register_buffer('attention_bias_idxs', torch.ones(49, 49).long())

        #     self.attention_biases_seg = torch.nn.Parameter(
        #         torch.zeros(self.num_attention_heads, len(attention_offsets)))
        #     self.register_buffer('attention_bias_idxs_seg',
        #                          torch.LongTensor(idxs).view(N, N))

        self.use_cross_attention = use_cross_attention

    def transpose_for_scores(self, x: Tensor, attention_head_size: int) -> Tensor:
        new_x_shape = x.size()[:-1] + (self.num_attention_heads, attention_head_size)
        x = x.view(new_x_shape)
        return x.permute(0, 2, 1, 3)

    def sequence_reduce(self, x: Tensor, height: int, width: int) -> Tensor:
        """SegFormer
        """
        B, N, C = x.shape
        # Reshape to (batch_size, num_channels, height, width)
        x = x.permute(0, 2, 1).reshape(B, C, height, width)
        # Apply sequence reduction
        x = self.sr(x)
        # Reshape back to (batch_size, seq_len, num_channels)
        x = x.reshape(B, C, -1).permute(0, 2, 1)
        x = self.sr_layer_norm(x)
        return x

    def forward(
        self,
        query_states: Tensor,
        key_value_states: Optional[Tensor] = None,
        head_mask: Optional[Tensor] = None,
        height: Optional[int] = None,
        width: Optional[int] = None
    ) -> Union[Tuple[Tensor, Tensor], Tuple[Tensor]]:
        """Forward pass for multi-head attention (self- or cross-)

        Let. S_s(source): S_q(query) (= H'*W' + 1)
        Let. S_t(target): S_k(key) = S_v(value)
        If self-attention, S_s = S_t
        Let. C_qk: C_qksplit * {head}(=num_attention_heads) = (mostly) hidden_size
        Let. C_v: C_vsplit * {head}(=num_attention_heads) * attn_ratio = (mostly) hidden_size * attn_ratio
        Mostly, C = C_qk = C_v (in_channels = hidden_size = value_hidden_size)
        Args:
            query_states (Tensor): [B x S_s x C]
            key_value_states (Optional[Tensor], optional): [B x S_t x C]. Defaults to None.
            head_mask (Optional[Tensor], optional): [B x {head} x S_s x S_t]. Defaults to None.
            height (Optional[int], optional): (Segformer only). Defaults to None.
            width (Optional[int], optional): (Segformer only). Defaults to None.

        Returns:
            Union[Tuple[Tensor, Tensor], Tuple[Tensor]]: output tensor or list of attention matrices
        """

        mixed_query_layer = self.query(query_states)  # B x S_s x C_qk

        if not self.use_cross_attention:  # Self-attention
            key_value_states = query_states  # B x S_t(=S_s) x C_qk
        if self.use_sequence_reduction:
            key_value_states = self.sequence_reduce(key_value_states, height, width)  # B x S_t' x C_qk

        key_layer = self.transpose_for_scores(self.key(key_value_states), self.attention_head_size)  # B x {head} x S_t x C_qksplit
        value_layer = self.transpose_for_scores(self.value(key_value_states), self.value_attention_head_size)  # B x {head} x S_t x C_vsplit
        query_layer = self.transpose_for_scores(mixed_query_layer, self.attention_head_size)  # B x {head} x S_s x C_qksplit

        attention_scores = torch.matmul(query_layer, key_layer.transpose(-1, -2))  # B x {head} x S_s x S_t

        attention_scores = attention_scores / self.attention_scale  # B x {head} x S_s x S_t

        if self.use_attention_bias:
            bias = self.attention_biases[:, self.attention_bias_idxs]
            bias = nn.functional.interpolate(bias.unsqueeze(0), size=(attention_scores.size(-2), attention_scores.size(-1)), mode='bilinear')
            attention_scores = attention_scores + bias

        # Normalize the attention scores to probabilities.
        attention_probs = nn.functional.softmax(attention_scores, dim=-1)  # B x {head} x S_s x S_t

        attention_probs = self.dropout(attention_probs)  # B x {head} x S_s x S_t

        if head_mask is not None:
            attention_probs = attention_probs * head_mask  # B x {head} x S_s x S_t

        context_layer = torch.matmul(attention_probs, value_layer)  # B x {head} x S_s x C_vsplit

        context_layer = context_layer.permute(0, 2, 1, 3).contiguous()  # B x S_s x {head} x C_vsplit
        new_context_layer_shape = context_layer.size()[:-2] + (self.value_head_size,)
        context_layer = context_layer.view(new_context_layer_shape)  # B x S_s x C_v

        context_layer = self.linear(context_layer)  # B x S_s x C
        context_layer = self.dropout(context_layer)  # B x S_s x C

        if self.output_with_attentions:
            return (context_layer, attention_probs)

        return context_layer  # B x S_s x C


class MultiQueryAttention2D(nn.Module):
    def __init__(
        self,
        hidden_size,
        num_attention_heads,
        attention_hidden_size = None,
        value_hidden_size = None,
        attention_scale = None,
        attention_dropout_prob = 0.0,
        use_qkv_bias = True,
        use_cross_attention = False,
        output_with_attentions = False,
        query_pooling_stride: Optional[int] = None,
        key_val_downsample: Optional[bool] = None,
        key_val_downsample_kernel_size: Optional[int] = None,
        key_val_downsample_stride: Optional[int] = None,
        norm_type: Optional[str] = None,
    ) -> None:
        super().__init__()

        norm_type = norm_type if norm_type is not None else 'batch_norm'
        norm_layer = NORM_REGISTRY[norm_type]

        self.attention_hidden_size = attention_hidden_size if attention_hidden_size is not None else hidden_size
        if attention_hidden_size % num_attention_heads != 0:
            raise ValueError(
                f"The hidden size {attention_hidden_size,} is not a multiple of the number of attention "
                f"heads {num_attention_heads}."
            )

        self.value_hidden_size = value_hidden_size if value_hidden_size is not None else int(self.attention_hidden_size / num_attention_heads)

        self.num_attention_heads = num_attention_heads
        self.attention_head_size = int(attention_hidden_size / num_attention_heads)

        self.attention_scale = attention_scale if attention_scale is not None \
            else math.sqrt(self.attention_head_size)

        self.query = []
        if query_pooling_stride:
            self.query.append(Pool2d(method='avg', kernel_size=query_pooling_stride))
            self.query.append(norm_layer(hidden_size))
        self.query.append(nn.Conv2d(hidden_size, self.attention_hidden_size,
                                    kernel_size=1, bias=use_qkv_bias))
        self.query = nn.Sequential(*self.query)

        self.key = []
        self.value = []
        if key_val_downsample:
            self.key.append(nn.Conv2d(hidden_size, hidden_size, kernel_size=key_val_downsample_kernel_size, padding=key_val_downsample_kernel_size // 2,
                                      stride=key_val_downsample_stride, bias=use_qkv_bias, groups=hidden_size))
            self.key.append(norm_layer(hidden_size))
            self.value.append(nn.Conv2d(hidden_size, hidden_size, kernel_size=key_val_downsample_kernel_size, padding=key_val_downsample_kernel_size // 2,
                                        stride=key_val_downsample_stride, bias=use_qkv_bias, groups=hidden_size))
            self.value.append(norm_layer(hidden_size))

        self.key.append(nn.Conv2d(hidden_size, self.attention_head_size, kernel_size=1, bias=use_qkv_bias))
        self.value.append(nn.Conv2d(hidden_size, self.value_hidden_size, kernel_size=1, bias=use_qkv_bias))
        self.key = nn.Sequential(*self.key)
        self.value = nn.Sequential(*self.value)

        self.linear = []
        if query_pooling_stride:
            self.linear.append(nn.Upsample(scale_factor=query_pooling_stride, mode='bilinear', align_corners=False))
        self.linear.append(nn.Conv2d(self.attention_hidden_size, hidden_size, kernel_size=1, bias=False))
        self.linear = nn.Sequential(*self.linear)

        self.dropout = nn.Dropout(attention_dropout_prob)
        self.output_with_attentions = output_with_attentions

        self.use_cross_attention = use_cross_attention

    def transpose_for_scores(self, x: Tensor, num_head: int) -> Tensor:
        x = x.flatten(-2)  # B x C x H x W -> B x C x {H * W}
        new_x_shape = (x.shape[0], num_head,  x.shape[1] // num_head, x.shape[2])
        x = x.view(new_x_shape)  # B x C x {H * W} -> B x {head} x C_split x {H * W}
        return x.permute(0, 1, 3, 2)  # B x {head} x C_split x {H * W} -> B x {head} x {H * W} x C_split

    def forward(
        self,
        query_states: Tensor,
        key_value_states: Optional[Tensor] = None,
        head_mask: Optional[Tensor] = None,
    ) -> Union[Tuple[Tensor, Tensor], Tuple[Tensor]]:
        B, _, H_q, W_q = query_states.shape  # B x hiddin_dim x H_q x W_q

        query = self.query(query_states)

        if not self.use_cross_attention:  # Self-attention
            key_value_states = query_states

        key = self.key(key_value_states)
        value = self.value(key_value_states)

        # query_states: B x C_q x H_q x W_q
        # key_value_states: B x C_kv x H_kv x W_kv

        key_layer = self.transpose_for_scores(key, 1)  # B x C_kv x H_kv x W_kv -> B x 1 x {H_kv * W_kv} x C_kv
        value_layer = self.transpose_for_scores(value, 1)  # B x C_kv x H_kv x W_kv -> B x 1 x {H_kv * W_kv} x C_kv
        query_layer = self.transpose_for_scores(query, self.num_attention_heads)  # B x C_q x H_q x W_q -> B x {head} x {H_q * W_q} x C_qsplit

        attention_scores = torch.matmul(query_layer, key_layer.transpose(-1, -2))

        attention_scores = attention_scores / self.attention_scale

        # Normalize the attention scores to probabilities.
        attention_probs = nn.functional.softmax(attention_scores, dim=-1)

        attention_probs = self.dropout(attention_probs)

        if head_mask is not None:
            attention_probs = attention_probs * head_mask

        context_layer = torch.matmul(attention_probs, value_layer)

        context_layer = context_layer.permute(0, 1, 3, 2).contiguous()  # B x {head} x {H_q * W*q} x C_qsplit -> B x {head} x C_qsplit x {H_q * W*q}
        context_layer = context_layer.reshape(B, self.attention_hidden_size, H_q, W_q)  # B x {head} x C_qsplit x {H_q * W*q} -> B x C_q x H_q x W*q

        context_layer = self.linear(context_layer)  # B x C_q x H_q x W_q -> B x hiddin_dim x H_q x W_q
        context_layer = self.dropout(context_layer)

        if self.output_with_attentions:
            return (context_layer, attention_probs)

        return context_layer  # B x hiddin_dim x H_q x W_


class MobileMultiQueryAttention2D(nn.Module):
    def __init__(
        self,
        hidden_size,
        num_attention_heads,
        attention_hidden_size = None,
        value_hidden_size = None,
        attention_scale = None,
        attention_dropout_prob = 0.0,
        use_qkv_bias = True,
        use_cross_attention = False,
        output_with_attentions = False,
        query_pooling_stride: Optional[int] = None,
        key_val_downsample: Optional[bool] = None,
        key_val_downsample_kernel_size: Optional[int] = None,
        key_val_downsample_stride: Optional[int] = None,
        norm_type: Optional[str] = None,
        layer_scale: Optional[float] = None,
        drop_path: float = 0.0,
    ) -> None:
        super().__init__()
        norm_type = norm_type if norm_type is not None else 'batch_norm'
        self.norm_layer = NORM_REGISTRY[norm_type](hidden_size)
        self.attention = MultiQueryAttention2D(
            hidden_size, num_attention_heads, attention_hidden_size, value_hidden_size, attention_scale, attention_dropout_prob,
            use_qkv_bias, use_cross_attention, output_with_attentions, query_pooling_stride, key_val_downsample,
            key_val_downsample_kernel_size, key_val_downsample_stride, norm_type
        )
        self.apply_layer_scale = False
        if layer_scale is not None:
            self.apply_layer_scale = True
            self.layer_scale = LayerScale2d(hidden_size, layer_scale)
        self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()

    def forward(
        self,
        query_states: Tensor,
        key_value_states: Optional[Tensor] = None,
        head_mask: Optional[Tensor] = None,
    ) -> Union[Tuple[Tensor, Tensor], Tuple[Tensor]]:
        residual = query_states

        out = self.norm_layer(query_states)
        out = self.attention(out, key_value_states, head_mask)
        if self.apply_layer_scale:
            out = self.layer_scale(out)
        out = self.drop_path(out) + residual
        return out


class ChannelMLP(nn.Module):
    def __init__(self, hidden_size, intermediate_size, hidden_dropout_prob, hidden_activation_type='silu'):
        super().__init__()
        self.ffn = nn.Sequential()
        self.ffn.add_module('dense1', nn.Linear(in_features=hidden_size, out_features=intermediate_size, bias=True))
        self.ffn.add_module('act', ACTIVATION_REGISTRY[hidden_activation_type]())
        self.ffn.add_module('dropout', nn.Dropout(p=hidden_dropout_prob))
        self.ffn.add_module('dense2', nn.Linear(in_features=intermediate_size, out_features=hidden_size, bias=True))

        self.dropout = nn.Dropout(p=hidden_dropout_prob)

    def forward(self, x):
        x = self.ffn(x)
        x = self.dropout(x)
        return x

class MetaFormerBlock(nn.Module):
    def __init__(self, hidden_size, layer_norm_eps) -> None:
        super().__init__()
        self.layernorm_before = nn.LayerNorm(hidden_size)
        self.layernorm_after = nn.LayerNorm(hidden_size)
        self.token_mixer = nn.Identity()  # MultiHeadAttention()
        self.channel_mlp = nn.Identity()  # ChannelMLP()

    def forward(self, x):
        out_token_mixer = self.layernorm_before(x)
        out_token_mixer = self.token_mixer(out_token_mixer)

        out_token_mixer = out_token_mixer + x

        out_final = self.layernorm_after(out_token_mixer)
        out_final = self.channel_mlp(out_final)

        out_final = out_final + out_token_mixer

        return out_final

class MetaFormerEncoder(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.blocks = nn.Identity()
        # self.blocks = nn.Sequential(
        #     *[MetaFormerBlock(hidden_size, layer_norm_eps) for _ in range(num_layers)]
        # )

    def forward(self, x):
        x = self.blocks(x)
        return x

class MetaFormer(nn.Module):
    def __init__(self, hidden_sizes) -> None:
        super().__init__()
        self._feature_dim = hidden_sizes[-1]
        self._intermediate_features_dim = hidden_sizes

        self.patch_embed = nn.Identity()
        self.encoder = MetaFormerEncoder()
        self.norm = nn.Identity()

    @property
    def feature_dim(self):
        return self._feature_dim

    @property
    def intermediate_features_dim(self):
        return self._intermediate_features_dim

    def forward(self, x: FXTensorType):
        x = self.patch_embed(x)
        x = self.encoder(x)
        x = self.norm(x)
        feat = torch.mean(x, dim=1)
        return BackboneOutput(last_feature=feat)



# Copyright (C) 2024 Nota Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ----------------------------------------------------------------------------

from typing import Dict, Type

from torch.optim.optimizer import Optimizer

from .custom import SGD, Adadelta, Adagrad, Adam, Adamax, AdamW, RMSprop

OPTIMIZER_DICT: Dict[str, Type[Optimizer]] = {
    'adamw': AdamW,
    'adam': Adam,
    'adadelta': Adadelta,
    'adagrad': Adagrad,
    'rmsprop': RMSprop,
    'adamax': Adamax,
    'sgd': SGD,
}
